# -*- coding: utf-8 -*-
"""
Define functions useful for conversion between different NFW conventions

This module contains the functions used to move between NFW conventions and
to transform NFW parameters into lenstronomy inputs.
"""
import numpy as np
from ..Utils import cosmology_utils
import numba
from scipy.interpolate import interp1d
import lenstronomy.Util.constants as const


@numba.njit()
def cored_nfw_integral(r_tidal,rho_nfw,r_scale,r_upper):
	"""Integrates the cored nfw mass profile from 0 to r_upper

	Args:
		r_tidal (float): The tidal radius within which the NFW profile
			will be replaced by a uniform profile. Units of kpc
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius. Units of M_sun/kpc^3
		r_scale (float): The scale radius of the nfw in units of kpc
		r_upper (np.array): An array containing the upper bounds
			to be evaluated in units of kpc.

	Returns:
		(np.array): The value of the integral for each r_upper given.

	Notes:
		The cored NFW profile here assume uniform distribution within the
		tidal radius and then an NFW radial distribution outside the
		tidal radius.
	"""
	# Convert to natural units for NFW
	x_tidal = r_tidal / r_scale
	x_upper = r_upper / r_scale

	# Get the value of the NFW in the core region such that at x_tidal we
	# match the nfw profile.
	linear_scaling = rho_nfw/(x_tidal*(1+x_tidal)**2)

	# Array to save the integral outputs to
	integral_values = np.zeros(r_upper.shape)

	# Add the cored component
	integral_values += 1.0/3.0*linear_scaling * np.minimum(r_tidal,r_upper)**3

	# Add the nfw component where x_upper > x_tidal
	lower_bound = 1/(x_tidal+1) + np.log(x_tidal+1)
	upper_bound = 1/(x_upper+1) + np.log(x_upper+1)
	nfw_integral = upper_bound - lower_bound
	add_nfw = r_upper > r_tidal
	integral_values[add_nfw] += nfw_integral[add_nfw]*rho_nfw*r_scale**3

	# 4 pi factor from volume integral.
	integral_values *= 4 * np.pi

	return integral_values


def cored_nfw_draws(r_tidal,rho_nfw,r_scale,r_max,n_subs,n_cdf_samps=1000):
	"""Returns radial samples from a cored nfw mass profile

	Args:
		r_tidal (float): The tidal radius within which the NFW profile
			will be replaced by a uniform profile in units of kpc.
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius in units of M_sun / kpc^3.
		r_scale (float): The scale radius of the nfw in units of kpc.
		r_max (float): The maximum value of r to sample in units of kpc.
		n_subs (int): The number of subhalo positions to sample
		n_cdf_samps (int): The number of samples to use to numerically
			invert the cdf for sampling.

	Returns:
		(np.array): A n_subs array giving sampled radii in units of kpc.

	Notes:
		The cored NFW profile here assume uniform distribution within the
		tidal radius and then an NFW radial distribution outside the
		tidal radius.
	"""
	# First we have to numerically calculate the inverse cdf
	r_values = np.linspace(0,r_max,n_cdf_samps)
	cdf_values = cored_nfw_integral(r_tidal,rho_nfw,r_scale,r_values)
	# Normalize
	cdf_values /= np.max(cdf_values)
	# Use scipy to create our inverse cdf
	inverse_cdf = interp1d(cdf_values,r_values)

	# Now draw from the inverse cdf
	cdf_draws = np.random.rand(n_subs)
	r_draws = inverse_cdf(cdf_draws)

	return r_draws


@numba.njit()
def nfw_integral(rho_nfw,r_scale,r_upper):
	"""Integrates the nfw mass profile from 0 to r_upper.

	Args:
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius. Units of M_sun/kpc^3
		r_scale (float): The scale radius of the nfw in units of kpc
		r_upper (np.array): An array containing the upper bounds
			to be evaluated in units of kpc.

	Returns:
		(np.array): The value of the integral for each r_upper given.
	"""
	# Convert to natural units for NFW
	x_upper = r_upper / r_scale

	# Calculate the nfw integral
	integral_values=1/(x_upper+1) + np.log(x_upper+1)-1

	# Deal with the change of units to x in the integral and 4 pi from the
	# volume element.
	return integral_values*rho_nfw*r_scale**3*4*np.pi


def nfw_draws(rho_nfw,r_scale,r_max,n_subs,n_cdf_samps=1000):
	"""Returns radial samples from the nfw mass profile.


	Args:
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius in units of M_sun / kpc^3.
		r_scale (float): The scale radius of the nfw in units of kpc.
		r_max (float): The maximum value of r to sample in units of kpc.
		n_subs (int): The number of subhalo positions to sample
		n_cdf_samps (int): The number of samples to use to numerically
			invert the cdf for sampling.

	Returns:
		(np.array): A n_subs array giving sampled radii in units of kpc.
	"""
	# First we have to numerically calculate the inverse cdf
	r_values = np.linspace(0,r_max,n_cdf_samps)
	cdf_values = nfw_integral(rho_nfw,r_scale,r_values)
	# Normalize
	cdf_values /= np.max(cdf_values)
	# Use scipy to create our inverse cdf
	inverse_cdf = interp1d(cdf_values,r_values)

	# Now draw from the inverse cdf
	cdf_draws = np.random.rand(n_subs)
	r_draws = inverse_cdf(cdf_draws)

	return r_draws


@numba.njit()
def tnfw_integral(rho_nfw,r_scale,r_trunc,r_upper):
	"""Integrates the truncated nfw mass profile from 0 to r_upper. This is
	the same as the nfw profile but scaled by a factor of
	r_trunc^2/(r_trunc^2+r^2)

	Args:
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius. Units of M_sun/kpc^3
		r_scale (float): The scale radius of the nfw in units of kpc
		r_trunc (float): The truncation radius for the profile.
		r_upper (np.array): An array containing the upper bounds
			to be evaluated in units of kpc.

	Returns:
		(np.array): The value of the integral for each r_upper given.
	"""
	# Do the integral calculation in pieces
	integral_values = 1/(2*(r_upper+r_scale)*(r_scale**2+r_trunc**2)**2)
	integral_values *= r_scale**3 * r_trunc**2

	integral_values *= (-2*r_upper*(r_scale**2+r_trunc**2)+
		(r_upper+r_scale)*(4*r_scale*r_trunc*np.arctan(r_upper/r_trunc)+
			(r_scale-r_trunc)*(r_scale+r_trunc)*(
				2*np.log(r_scale)-2*(
					np.log(r_upper+r_scale)+np.log(r_trunc))+np.log(
					r_upper**2+r_trunc**2))))

	return integral_values*4*np.pi*rho_nfw


def tnfw_draws(rho_nfw,r_scale,r_trunc,r_max,n_subs,n_cdf_samps=1000):
	"""Returns radial samples from truncated nfw mass profile This is
	the same as the nfw profile but scaled by a factor of
	r_trunc^2/(r_trunc^2+r^2)


	Args:
		rho_nfw (float): The amplitude of the nfw density outside the
			cored radius in units of M_sun / kpc^3.
		r_scale (float): The scale radius of the nfw in units of kpc.
		r_trunc (float): The truncation radius for the profile.
		r_max (float): The maximum value of r to sample in units of kpc.
		n_subs (int): The number of subhalo positions to sample
		n_cdf_samps (int): The number of samples to use to numerically
			invert the cdf for sampling.

	Returns:
		(np.array): A n_subs array giving sampled radii in units of kpc.

	Notes:
		Set r_max to a value larger than r_trunc but not so large
		that you're wasting a lot of interpolation points for the inverse
		cdf.
	"""
	# First we have to numerically calculate the inverse cdf
	r_values = np.linspace(0,r_max,n_cdf_samps)
	cdf_values = tnfw_integral(rho_nfw,r_scale,r_trunc,r_values)
	# Normalize
	cdf_values /= np.max(cdf_values)
	# Use scipy to create our inverse cdf
	inverse_cdf = interp1d(cdf_values,r_values)

	# Now draw from the inverse cdf
	cdf_draws = np.random.rand(n_subs)
	r_draws = inverse_cdf(cdf_draws)

	return r_draws


def r_200_from_m(m_200,z,cosmo):
	"""Calculates r_200 for our NFW profile given our m_200 mass.

	Args:
		m_200 (float, or np.array): The mass contained within r_200 in units
			of M_sun
		z (float, or np.array): The redshift of the halo
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		(float, or np.array): The r_200 radius corresponding to the given mass
		in units of kpc.

	Notes:
		This equation assumes that for a halo at redshift z, m200 is
		defined in terms of the critical density at that redshift. Therefore
		the output to the equation is in physical units.
	"""
	# Get the critical density
	h = cosmo.h
	# rho_c is returned in units of M_sun*h^2/kpc^3
	rho_c = cosmo.rho_c(z)*h**2

	# Return r_200 given that critical density
	return (3*m_200/(4*np.pi*rho_c*200))**(1.0/3.0)


def rho_nfw_from_m_c(m_200,c,cosmo,r_scale=None,z=None):
	"""Calculates the amplitude of the nfw profile given the physical
	parameters.

	Args:
		m_200 (float, or np.array): The mass of the nfw halo in units of M_sun
		c (float, or np.array): The concentration of the nfw_halo
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.
		r_scale (float, or np.array): The scale radius in units of kpc
		z (float, or np.array): The redshift at which to calculate
			r_200.

	Returns:
		(float, or np.array): The amplitude for the nfw in units of
		M_sun/kpc^3.
	"""
	# If r_scale is not provided, calculate it
	if r_scale is None:
		if z is None:
			raise ValueError('Must specify z if not specifying r_scale')
		r_200 = r_200_from_m(m_200,z,cosmo)
		r_scale = r_200/c

	# Calculate the density to match the mass and concentration.
	rho_0 = m_200/(4*np.pi*r_scale**3*(np.log(1+c)-c/(1+c)))

	return rho_0


def m_c_from_rho_r_scale(rho_nfw,r_scale,cosmo,z):
	"""Calculates the mass and concentration from the amplitude of the
	nfw profile and scale radius.

	Args:
		rho_nfw (np.array): The amplitude of the nfw in units of M_sun/kpc^3
		r_scale (np.array): The scale radius in units of kpc
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.
		z (float): The redshift of the NFW halos.

	Returns:
		[np.array,...]: A list of 2 np.array objects. This first is the
		mass and the second the concentration for each NFW.
	"""
	# The equations do not have a simple closed form, so we're going to
	# cheat and interpolate
	c_eval = np.logspace(-1,3,1000)
	h = cosmo.h
	# rho_c is returned in units of M_sun*h^2/kpc^3
	rho_c = cosmo.rho_c(z)*h**2

	rho_nfw_evals = rho_c*200/3*c_eval**3/(np.log(1+c_eval)-c_eval/(1+c_eval))
	# Interpolate the inverse
	inverse_c = interp1d(rho_nfw_evals,c_eval)
	c = inverse_c(rho_nfw)

	# With the concentration we can get the mass
	r_200 = r_scale * c
	m = (4*np.pi*rho_c*200*r_200**3)/3

	return [m,c]


def calculate_sigma_crit(z,z_source,cosmo):
	"""Calculates the critical density for the given configuration

	Args:
		z (np.array): The redshift of the nfw
		z_source (float): The redshift of the source
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		(np.array): The critical surface density in units of M_sun/kpc^2
	"""
	# Normalization factor for sigma_crit
	norm = const.c ** 2 / (4 * np.pi * const.G)*const.Mpc / const.M_sun
	# Get our three angular diameter distances
	mpc2kpc = 1e3
	dd = cosmo.angularDiameterDistance(z)
	ds = cosmo.angularDiameterDistance(z_source)
	cosmo_astro = cosmo.toAstropy()
	dds = cosmo_astro.angular_diameter_distance_z1z2(z,z_source).value
	return ds / (dd * dds) * norm/mpc2kpc**2


def convert_to_lenstronomy_NFW(r_scale,z,rho_nfw,z_source,cosmo):
	"""Converts physical NFW parameters to parameters used by lenstronomy

	Args:
		r_scale (np.array): The scale radius of the nfw in units of kpc
		z (np.array): The redshift of the nfw
		rho_nfw (np.array): The amplitude of the nfw halos in units of
			M_sun/kpc^3
		z_source (float): The redshift of the source
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		[np.array,...]: A list of 2 numpy arrays: The angular r_scale and the
		deflection angle at the scale radius.
	"""
	kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,cosmo)

	# For two of our parameters we just need to convert to arcseconds
	r_scale_ang = r_scale/kpc_per_arcsecond

	# For our deflection angle the calculation is a little more involved
	# and requires knowledge of the source redshift.
	sigma_crit = calculate_sigma_crit(z,z_source,cosmo)
	alpha_rs = rho_nfw * (4*r_scale**2*(1+np.log(0.5)))
	alpha_rs /= kpc_per_arcsecond * sigma_crit
	return [r_scale_ang, alpha_rs]


def convert_from_lenstronomy_NFW(r_scale_ang,alpha_rs,z,z_source,cosmo):
	"""Converts parameters used by lenstronomy to physical NFW parameters

	Args:
		r_scale_ang (np.array): The scale radius of the nfw in angular
			units
		alpha_rs (np.array): The deflection angle at the scale radius
		z (np.array): The redshift of the nfw
		z_source (float): The redshift of the source
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		[np.array,...]: A list of 2 numpy arrays: The scale radius in kpc
		and the amplitude of the nfw halo in units of M_sun/kpc^3.
	"""
	kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,cosmo)

	# For the scale radius we just need to convert from arcseconds
	r_scale = r_scale_ang * kpc_per_arcsecond

	# Deflection angle to rho_nfw requires sigma_crit and source
	# position
	sigma_crit = calculate_sigma_crit(z,z_source,cosmo)
	rho_nfw = alpha_rs / (4*r_scale**2*(1+np.log(0.5)))
	rho_nfw *= kpc_per_arcsecond * sigma_crit
	return [r_scale,rho_nfw]


def convert_to_lenstronomy_tNFW(r_scale,z,rho_nfw,r_trunc,z_source,cosmo):
	"""Converts physical tNFW parameters to parameters used by lenstronomy

	Args:
		r_scale (np.array): The scale radius of the nfw in units of kpc
		z (np.array): The redshift of the nfw
		rho_nfw (np.array): The amplitude of the nfw halos in units of
			M_sun/kpc^3
		r_trunc (np.array): The truncation radius for each nfw in units of
			kpc.
		z_source (float): The redshift of the source
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		[np.array,...]: A list of 3 numpy arrays: The angular r_scale, the
		deflection angle at the scale radius, and the angular truncation
		radius in units of kpc.
	"""
	kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,cosmo)

	# For two of our parameters we just need to convert to arcseconds
	r_scale_ang = r_scale/kpc_per_arcsecond
	r_trunc_ang = r_trunc/kpc_per_arcsecond

	# For our deflection angle the calculation is a little more involved
	# and requires knowledge of the source redshift. We will take advantage
	# of lenstronomy for this calculation.
	sigma_crit = calculate_sigma_crit(z,z_source,cosmo)
	alpha_rs = rho_nfw * (4*r_scale**2*(1+np.log(0.5)))
	alpha_rs /= kpc_per_arcsecond * sigma_crit
	return [r_scale_ang, alpha_rs, r_trunc_ang]


def convert_from_lenstronomy_tNFW(r_scale_ang,alpha_rs,r_trunc_ang,z,
	z_source,cosmo):
	"""Converts parameters used by lenstronomy to physical tNFW parameters

	Args:
		r_scale_ang (np.array): The scale radius of the nfw in angular
			units
		alpha_rs (np.array): The deflection angle at the scale radius
		r_trunc_ang (np.array): The truncation radius for each nfw in angular
			units.
		z (np.array): The redshift of the nfw
		z_source (float): The redshift of the source
		cosmo (colossus.cosmology.cosmology.Cosmology): An instance of the
			colossus cosmology object.

	Returns:
		[np.array,...]: A list of 3 numpy arrays: The scale radius in kpc,
		the amplitude of the nfw halo in units of M_sun/kpc^3, and the
		truncation radius in units of kpc.
	"""
	kpc_per_arcsecond = cosmology_utils.kpc_per_arcsecond(z,cosmo)

	# For the scale radius we just need to convert from arcseconds
	r_scale = r_scale_ang * kpc_per_arcsecond
	r_trunc = r_trunc_ang * kpc_per_arcsecond

	# Deflection angle to rho_nfw requires sigma_crit and source
	# position
	sigma_crit = calculate_sigma_crit(z,z_source,cosmo)
	rho_nfw = alpha_rs / (4*r_scale**2*(1+np.log(0.5)))
	rho_nfw *= kpc_per_arcsecond * sigma_crit
	return [r_scale,rho_nfw,r_trunc]



# -*- coding: utf-8 -*-
"""
Provides classes for specifying a sersic light distribution.

This module contains the class required to provide a sersic light distribution
as the source for paltas.
"""
from .source_base import SourceBase
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Util.data_util import magnitude2cps


class SingleSersicSource(SourceBase):
	"""Class to generate single Sersic profile light models

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters: dictionary with source-specific parameters.

	Notes:

	Required Parameters

	- magnitude - AB absolute magnitude of the source
	- output_ab_zeropoint - AB magnitude zeropoint of the detector
	- R_sersic - Sersic radius in units of arcseconds
	- n_sersic - Sersic index
	- e1 - x-direction ellipticity eccentricity
	- e2 - xy-direction ellipticity eccentricity
	- center_x - x-coordinate lens center in units of arcseconds
	- center_y - y-coordinate lens center in units of arcseconds
	- z_source - source redshift
	"""

	required_parameters = ('magnitude', 'output_ab_zeropoint', 'R_sersic', 
		'n_sersic', 'e1', 'e2', 'center_x', 'center_y', 'z_source')

	def draw_source(self):
		"""Return lenstronomy LightModel kwargs

		Returns:
			(list,list) A list containing the model names(s), and
			a list containing the model kwargs dictionaries.
		"""
		# Just extract each of the sersic parameters.
		sersic_params ={
			k: v
			for k, v in self.source_parameters.items()
			if k in self.required_parameters}
		sersic_params.pop('z_source')
		sersic_params.pop('output_ab_zeropoint')

		# mag to amp conversion
		sersic_params.pop('magnitude')
		sersic_params['amp'] = SingleSersicSource.mag_to_amplitude(
			self.source_parameters['magnitude'],
			self.source_parameters['output_ab_zeropoint'], sersic_params)
		return (
			['SERSIC_ELLIPSE'],
			[sersic_params])

	@staticmethod
	def mag_to_amplitude(mag, mag_zero_point, kwargs_list):
		"""Converts a user defined magnitude to the corresponding amplitude
		that lenstronomy will use
	
		Args:
			mag (float): user defined desired magnitude
			kwargs_list (dict): dict of kwargs for SERSIC_ELLIPSE, amp 
			parameter not included

		Returns: 
			(float): amplitude lenstronomy should use to get desired magnitude
			desired magnitude
		"""

		sersic_model = LightModel(['SERSIC_ELLIPSE'])
		# norm=True sets amplitude = 1
		flux_norm = sersic_model.total_flux([kwargs_list], norm=True)[0]
		flux_true = magnitude2cps(mag, mag_zero_point)
		
		return flux_true/flux_norm



# -*- coding: utf-8 -*-
"""
Generate simulated strong lensing images using the classes and parameters of
an input configuration dictionary.

This script generates strong lensing images from paltas config dictionaries.

Example
-------
To run this script, pass in the desired config as argument::

	$ python -m generate.py path/to/config.py path/to/save_folder --n 1000

The parameters will be pulled from config.py and the images will be saved in
save_folder. If save_folder doesn't exist it will be created.
"""
import numpy as np
import argparse, os, sys, warnings, copy
import shutil
from importlib import import_module
from paltas.Sampling.sampler import Sampler
from paltas.Utils.cosmology_utils import get_cosmology, ddt
from paltas.Utils.hubble_utils import hubblify
from paltas.Utils.lenstronomy_utils import PSFHelper
from paltas.Sources.galaxy_catalog import GalaxyCatalog
import matplotlib.pyplot as plt
from tqdm import tqdm
import pandas as pd
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.PointSource.point_source import PointSource
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.SimulationAPI.data_api import DataAPI
from lenstronomy.Data.psf import PSF
from lenstronomy.SimulationAPI.observation_api import SingleBand
import lenstronomy.Util.util as util

# Global filters on the python warnings. Using this since filter
# behaviour is a bit weird.
SERIALIZATIONWARNING = True
KWARGSNUMERICWARNING1 = True
KWARGSNUMERICWARNING2 = True


def parse_args():
	"""Parse the input arguments by the user

	Returns:
		(argparse.Namespace): An instance of the Namespace object with the
		users provided values.

	"""
	# Initialize the parser and the possible inputs
	parser = argparse.ArgumentParser()
	parser.add_argument('config_dict', help='Path to paltas configuration dict')
	parser.add_argument('save_folder', help='Folder to save images to')
	parser.add_argument('--n', default=1, dest='n', type=int,
		help='Size of dataset to generate (default 1)')
	parser.add_argument('--save_png_too', action='store_true',
		help='Also save a PNG for each image, for debugging')
	parser.add_argument('--tf_record', action='store_true',
		help='Generate the tf record for the training set.')
	args = parser.parse_args()
	return args


def draw_image(sample,los_class,subhalo_class,main_deflector_class,
	source_class,lens_light_class,point_source_class,numpix,multi_plane,
	kwargs_numerics,mag_cut,add_noise,apply_psf=True):
	""" Takes a sample from Sampler.sample toghether with the lenstronomy
	models and draws an image of the strong lensing system

	Args:
		sample (dict): A dictionary containing the parameter values that will
			be sampled.
		los_class (paltas.Substructure.LOSBase): An instance of the los class.
			If None no los will be added.
		subhalo_class (paltas.Substructure.SubhalosBase): An instance of the
			subhalo class. If None no subhalos will be added.
		main_deflector_class (paltas.MainDeflector.MainDeflectorBase): An
			instance of the main deflector class. If None no main deflector
			will be added.
		source_class (paltas.Sources.SourceBase): An instance of the
			SourceBase class that will be used to draw the source
			image.
		lens_light_class (paltas.Sources.SourceBase): An instance of the
			SourceBase class that will be used to draw the lens light. If
			None, no lens light will be added.
		point_source_class (paltas.PointSource.PointSourceBase): An instance of
			the PointSourceBase class will be used to draw a point source for
			the image. If None, no point source will be added.
		numpix (int): The number of pixels in the generated image
		multi_plane (bool): If true the multiplane approximation will
			be used. Must be true for los.
		kwargs_numerics (dict): A dict containing the numerics kwargs to
			pass to lenstronomy
		mag_cut (float): A magnification cut to enforce on the images. If
			None no magnitude cut will be enforced.
		add_noise (bool): If true noise will be added to the image. If False
			noise must be added after the fact.
		apply_psf (bool): If true the psf from sample will be applied. IF
			False the psf must be applied after the fact.

	Returns
		(np.array,dict): A numpy array containing the generated image
		and a metavalue dictionary with the corresponding sampled
		values.

	"""
	# Save the parameter values
	meta_values = {}

	# Pull our sampled parameters
	z_lens = sample['main_deflector_parameters']['z_lens']
	z_source = sample['source_parameters']['z_source']
	cosmo = get_cosmology(sample['cosmology_parameters'])

	# Populate the list of models and kwargs we need
	complete_lens_model_list = []
	complete_lens_model_kwargs = []
	complete_z_list = []
	lens_light_model_list = []
	lens_light_kwargs_list = []
	point_source_model_list = []
	point_source_kwargs_list = []

	# Get the remaining observation kwargs from the sampler and
	# initialize our observational objects.
	kwargs_psf = sample['psf_parameters']
	kwargs_detector = sample['detector_parameters']
	# Only apply a psf if required by the user.
	if apply_psf:
		psf_model = PSF(**kwargs_psf)
	else:
		psf_model = PSF(psf_type='NONE')
	data_api = DataAPI(numpix=numpix,**kwargs_detector)
	single_band = SingleBand(**kwargs_detector)

	# For each lensing object that's present, add them to the model and
	# kwargs list
	if los_class is not None:
		los_class.update_parameters(
			sample['los_parameters'],sample['main_deflector_parameters'],
			sample['source_parameters'],sample['cosmology_parameters'])
		los_model_list, los_kwargs_list, los_z_list = los_class.draw_los()
		interp_model_list, interp_kwargs_list, interp_z_list = (
			los_class.calculate_average_alpha(numpix*
				kwargs_numerics['supersampling_factor']))
		complete_lens_model_list += los_model_list + interp_model_list
		complete_lens_model_kwargs += los_kwargs_list + interp_kwargs_list
		complete_z_list += los_z_list + interp_z_list
	if subhalo_class is not None:
		subhalo_class.update_parameters(
			sample['subhalo_parameters'],
			sample['main_deflector_parameters'],
			sample['source_parameters'],sample['cosmology_parameters'])
		sub_model_list, sub_kwargs_list, sub_z_list = (
			subhalo_class.draw_subhalos())
		complete_lens_model_list += sub_model_list
		complete_lens_model_kwargs += sub_kwargs_list
		complete_z_list += sub_z_list
	if main_deflector_class is not None:
		main_deflector_class.update_parameters(
			sample['main_deflector_parameters'],sample['cosmology_parameters'])
		main_model_list, main_kwargs_list, main_z_list = (
			main_deflector_class.draw_main_deflector())
		complete_lens_model_list += main_model_list
		complete_lens_model_kwargs += main_kwargs_list
		complete_z_list += main_z_list
	if lens_light_class is not None:
		lens_light_class.update_parameters(
			cosmology_parameters=sample['cosmology_parameters'],
			source_parameters=sample['lens_light_parameters'])
		lens_light_model_list, lens_light_kwargs_list = (
			lens_light_class.draw_source())
	if point_source_class is not None:
		point_source_class.update_parameters(sample['point_source_parameters'])
		point_source_model_list,point_source_kwargs_list = (
			point_source_class.draw_point_source())

	complete_lens_model = LensModel(complete_lens_model_list, z_lens,
		z_source,complete_z_list, cosmo=cosmo.toAstropy(),
		multi_plane=multi_plane)

	# Now get the model and kwargs from the source class and turn it
	# into a source model
	source_class.update_parameters(
		cosmology_parameters=sample['cosmology_parameters'],
		source_parameters=sample['source_parameters'])
	# For catalog objects we also want to save the catalog index
	# and the (possibly randomized) additional rotation angle
	if isinstance(source_class,GalaxyCatalog):
		catalog_i, phi = source_class.fill_catalog_i_phi_defaults()
		meta_values['source_parameters_catalog_i'] = catalog_i
		meta_values['source_parameters_phi'] = phi
		source_model_list, source_kwargs_list = source_class.draw_source(
			catalog_i=catalog_i, phi=phi)
	else:
		source_model_list, source_kwargs_list = source_class.draw_source()
	source_light_model = LightModel(source_model_list)

	lens_light_model = LightModel(lens_light_model_list)

	# point source may need lens eqn solver kwargs
	lens_equation_params = None
	if 'lens_equation_solver_parameters' in sample.keys():
		lens_equation_params = sample['lens_equation_solver_parameters']
	point_source_model = PointSource(point_source_model_list,
		lensModel=complete_lens_model, save_cache=True, 
		kwargs_lens_eqn_solver=lens_equation_params)

	# Put it together into an image model
	complete_image_model = ImageModel(data_api.data_class,psf_model,
		complete_lens_model,source_light_model,lens_light_model,
		point_source_model,kwargs_numerics=kwargs_numerics)

	# Generate our image
	image = complete_image_model.image(complete_lens_model_kwargs,
		source_kwargs_list, lens_light_kwargs_list, point_source_kwargs_list)

	# Check for magnification cut and apply
	if mag_cut is not None:
		# Sum in case there's more than one source_light_model
		mag = np.sum(image)/np.sum(source_light_model.total_flux(
			source_kwargs_list))
		if mag < mag_cut:
			return None,None

	# If no noise is specified, don't add noise
	if add_noise:
		image += single_band.noise_for_model(image)

	# Calculate time delays and image positions, then add to meta_values
	if point_source_class is not None:
		# Calculate image positions
		x_image, y_image = point_source_model.image_position(
			point_source_kwargs_list,complete_lens_model_kwargs)
		num_images = len(x_image[0])
		pfix = 'point_source_parameters_'
		meta_values[pfix+'num_images'] = num_images

		# Calculate magnifications using complete_lens_model
		magnifications = complete_lens_model.magnification(x_image[0],
			y_image[0],complete_lens_model_kwargs)
		# If mag_pert is defined, add that pertubation
		if 'mag_pert' in sample['point_source_parameters'].keys():
			magnifications = magnifications * (
				sample['point_source_parameters']['mag_pert'][
					0:len(magnifications)])

		# Calculate time delays
		if sample['point_source_parameters']['compute_time_delays']:
			if 'kappa_ext' in sample['point_source_parameters'].keys():
				td = complete_lens_model.arrival_time(x_image[0],y_image[0],
					complete_lens_model_kwargs,
					sample['point_source_parameters']['kappa_ext'])
				# apply errors if defined in config_dict
				if 'time_delay_errors' in (
					sample['point_source_parameters'].keys()):
					errors = sample['point_source_parameters'][
						'time_delay_errors']
					errors = errors[0:len(td)-1]
					td = td + errors
				td -= td[0]
			else:
				raise ValueError('must define kappa_ext in point_source ' +
					'parameters to compute time delays')

			# TODO: calculate time delay distance
			meta_values[pfix+'ddt'] = ddt(sample,source_class.cosmo)

		for i in range(0,4):
			if i < num_images:
				meta_values[pfix+'x_image_'+str(i)] = x_image[0][i]
				meta_values[pfix+'y_image_'+str(i)] = y_image[0][i]
				meta_values[pfix+'magnification_'+str(i)] = magnifications[i]
			else:
				meta_values[pfix+'x_image_'+str(i)] = np.nan
				meta_values[pfix+'y_image_'+str(i)] = np.nan
				meta_values[pfix+'magnification_'+str(i)] = np.nan
				
			if sample['point_source_parameters']['compute_time_delays']:
				if i < len(td):
					meta_values[pfix+'time_delay_' + str(i)] = td[i]
				else:
					meta_values[pfix+'time_delay_' + str(i)] = np.nan

	return image,meta_values


def draw_drizzled_image(sample,los_class,subhalo_class,main_deflector_class,
	source_class,lens_light_class,point_source_class,numpix,multi_plane,
	kwargs_numerics,mag_cut,add_noise):
	""" Takes a sample from Sampler.sample toghether with the lenstronomy
	models and draws an image of the strong lensing system with a
	simulated drizzling pipeline applied.

	Args:
		sample (dict): A dictionary containing the parameter values that will
			be sampled.
		los_class (paltas.Substructure.LOSBase): An instance of the los class.
			If None no los will be added.
		subhalo_class (paltas.Substructure.SubhalosBase): An instance of the
			subhalo class. If None no subhalos will be added.
		main_deflector_class (paltas.MainDeflector.MainDeflectorBase): An
			instance of the main deflector class. If None no main deflector
			will be added.
		lens_light_class (paltas.Sources.SourceBase): An instance of the
			SourceBase class that will be used to draw the lens light. If
			None, no lens light will be added.
		lens_light_class (paltas.Sources.SourceBase): An instance of the
			SourceBase class that will be used to draw the lens light
		point_source_class (paltas.PointSource.PointSourceBase): An instance of
			the PointSourceBase class will be used to draw a point source for
			the image. If None, no point source will be added.
		numpix (int): The number of pixels in the generated image
		multi_plane (bool): If true the multiplane approximation will
			be used. Must be true for los.
		kwargs_numerics (dict): A dict containing the numerics kwargs to
			pass to lenstronomy
		mag_cut (float): A magnification cut to enforce on the images. If
			None no magnitude cut will be enforced.
		add_noise (bool): If true noise will be added to the image. If False
			noise must be added after the fact.
		drizzle_parameters (dict): The parameters to use for the drizzling
			operation that will be passed to hubblify.

	Returns
		(np.array,dict): A numpy array containing the generated image
		and a metavalue dictionary with the corresponding sampled
		values.
	"""
	# Grab our warning flags
	global KWARGSNUMERICWARNING1
	global KWARGSNUMERICWARNING2

	# Copy the sample since we will be modifying some parameters
	sample = copy.deepcopy(sample)

	# Generate a high resolution version of the image.
	supersample_pixel_scale = sample['drizzle_parameters'][
		'supersample_pixel_scale']
	output_pixel_scale = sample['drizzle_parameters']['output_pixel_scale']
	detector_pixel_scale = sample['detector_parameters']['pixel_scale']
	offset_pattern = sample['drizzle_parameters']['offset_pattern']
	wcs_distortion = sample['drizzle_parameters']['wcs_distortion']
	ss_scaling = detector_pixel_scale/supersample_pixel_scale
	numpix_ss = int(numpix*ss_scaling)

	# Temporairly reset the detector pixel_scale to the supersampled scale.
	sample['detector_parameters']['pixel_scale'] = supersample_pixel_scale

	# Modify the numerics kwargs to account for the supersampled pixel scale
	if 'supersampling_factor' in kwargs_numerics:
		if KWARGSNUMERICWARNING1:
			warnings.warn('kwargs_numerics supersampling_factor modified '
				+'for drizzle',category=RuntimeWarning)
			KWARGSNUMERICWARNING1 = False
		kwargs_numerics['supersampling_factor'] = max(1,
			int(kwargs_numerics['supersampling_factor']/ss_scaling))
	if 'point_source_supersampling_factor' in kwargs_numerics:
		if KWARGSNUMERICWARNING2:
			warnings.warn('kwargs_numerics point_source_supersampling_factor'
				+ ' modified for drizzle',category=RuntimeWarning)
			KWARGSNUMERICWARNING2 = False
		kwargs_numerics['point_source_supersampling_factor'] = max(1,int(
			kwargs_numerics['point_source_supersampling_factor']/ss_scaling))

	# Use the normal generation class to make our highres image without
	# noise.
	image_ss, meta_values = draw_image(sample,los_class,subhalo_class,
		main_deflector_class,source_class,lens_light_class,point_source_class, 
		numpix_ss,multi_plane,kwargs_numerics, mag_cut,False,apply_psf=False)
	sample['detector_parameters']['pixel_scale'] = detector_pixel_scale

	# Deal with an image that does not pass a cut.
	if image_ss is None:
		return image_ss, meta_values

	# Grab the PSF supersampling factor if present.
	if 'psf_supersample_factor' in sample['drizzle_parameters']:
		psf_supersample_factor = (
			sample['drizzle_parameters']['psf_supersample_factor'])
	else:
		warnings.warn('No psf_supersample_factor provided so 1 will be assumed.')
		psf_supersample_factor = 1

	if psf_supersample_factor > ss_scaling:
		raise ValueError(f'psf_supersample_factor {psf_supersample_factor} '
			+ f'larger than the supersampling {ss_scaling} defined in the '
			+'drizzle parameters.')

	# Make sure that if the user provided a PIXEL psf that the user specified
	# the same supersampling factor as for the drizzle_parameters
	if sample['psf_parameters']['psf_type'] == 'PIXEL' and (
		'point_source_supersampling_factor' not in sample['psf_parameters'] or
		sample['psf_parameters']['point_source_supersampling_factor'] !=
		psf_supersample_factor):
		raise ValueError('Must specify point_source_supersampling_factor for '
			'PIXEL psf and the value must equal psf_supersample_factor in the '
			' drizzle parameters')

	# We'll bypass lenstronomy's supersampling code by modifying
	# the data api.
	kwargs_numerics['point_source_supersampling_factor'] = 1
	kwargs_numerics['supersampling_factor'] = 1
	kwargs_numerics['supersampling_convolution'] = False
	sample['psf_parameters']['point_source_supersampling_factor'] = 1
	sample['detector_parameters']['pixel_scale'] /= psf_supersample_factor

	# Create our noise and psf models.
	kwargs_detector = sample['detector_parameters']
	kwargs_psf = sample['psf_parameters']
	single_band = SingleBand(**kwargs_detector)
	data_class = DataAPI(numpix=numpix*psf_supersample_factor,
		**kwargs_detector).data_class
	psf_model_lenstronomy = PSF(**kwargs_psf)

	if add_noise:
		noise_model = single_band.noise_for_model
	else:
		def noise_model(image):
			return 0

	psf_model = PSFHelper(data_class,psf_model_lenstronomy,
		kwargs_numerics).psf_model

	# Pass the high resolution image to hubblify
	image = hubblify(image_ss,supersample_pixel_scale,detector_pixel_scale,
		output_pixel_scale,noise_model,psf_model,offset_pattern,
		wcs_distortion=wcs_distortion,pixfrac=1.0,kernel='square',
		psf_supersample_factor=psf_supersample_factor)

	return image, meta_values


def main():
	"""Generates the strong lensing images by drawing parameters values from
	the provided configuration dictionary.
	"""
	# Get the user provided arguments
	args = parse_args()

	# Setup the warning filter
	global SERIALIZATIONWARNING

	# Get the dictionary from the provided .py file
	config_dir, config_file = os.path.split(os.path.abspath(args.config_dict))
	sys.path.insert(0, config_dir)
	config_name, _ = os.path.splitext(config_file)
	config_module = import_module(config_name)
	config_dict = config_module.config_dict

	# Set a random seed if provided
	if hasattr(config_module,'seed'):
		np.random.seed(config_module.seed)

	# Make the directory if not already there
	if not os.path.exists(args.save_folder):
		os.makedirs(args.save_folder)
	print("Save folder path: {:s}".format(args.save_folder))

	# Copy out config dict
	shutil.copy(
		os.path.abspath(args.config_dict),
		args.save_folder)

	# Set up our sampler and draw an initial sample for initialization
	sampler = Sampler(config_dict)
	sample = sampler.sample()

	# Get the numerical kwargs numpix from the config
	kwargs_numerics = config_module.kwargs_numerics
	numpix = config_module.numpix

	# Set up the paltas objects we'll use
	multi_plane = False
	los_class = None
	subhalo_class = None
	main_deflector_class = None
	lens_light_class = None
	point_source_class = None
	do_drizzle = False
	if 'los' in config_dict:
		los_class = config_dict['los']['class'](sample['los_parameters'],
			sample['main_deflector_parameters'],sample['source_parameters'],
			sample['cosmology_parameters'])
		multi_plane = True
	if 'subhalo' in config_dict:
		subhalo_class = config_dict['subhalo']['class'](
			sample['subhalo_parameters'],sample['main_deflector_parameters'],
			sample['source_parameters'],sample['cosmology_parameters'])
	if 'main_deflector' in config_dict:
		main_deflector_class = config_dict['main_deflector']['class'](
			sample['main_deflector_parameters'],sample['cosmology_parameters'])
	if 'drizzle' in config_dict:
		do_drizzle = True
	if 'lens_light' in config_dict:
		lens_light_class = config_dict['lens_light']['class'](
			sample['cosmology_parameters'], sample['lens_light_parameters'])
	if 'point_source' in config_dict:
		point_source_class = config_dict['point_source']['class'](
			sample['point_source_parameters'])

	source_class = config_dict['source']['class'](
		sample['cosmology_parameters'],sample['source_parameters'])

	# Use a pandas dataframe to store the parameter values.
	metadata = pd.DataFrame()
	metadata_path = os.path.join(args.save_folder,'metadata.csv')

	# See if a magnitude cut was specified
	if hasattr(config_module, 'mag_cut'):
		mag_cut = config_module.mag_cut
	else:
		mag_cut = None

	# See if ignoring noise was specified
	if hasattr(config_module,'no_noise'):
		add_noise = not config_module.no_noise
	else:
		add_noise = True

	# Generate our images
	pbar = tqdm(total=args.n)
	nt = 0
	tries = 0
	while nt < args.n:
		# We always try
		tries += 1

		# Draw a sample and generate the image.
		sample = sampler.sample()
		kwargs_detector = sample['detector_parameters']

		# Use the drizzling pipeline if the drizzling parameters were
		# provided.
		if do_drizzle:
			image,meta_values = draw_drizzled_image(sample,los_class,
				subhalo_class,main_deflector_class,source_class,lens_light_class,
				point_source_class,numpix,multi_plane,kwargs_numerics,
				mag_cut,add_noise)
		else:
			image,meta_values = draw_image(sample,los_class,subhalo_class,
				main_deflector_class,source_class,lens_light_class,
				point_source_class,numpix,multi_plane,kwargs_numerics,mag_cut,
				add_noise)

		# Failed attempt if there is no image output
		if image is None:
			continue

		# Mask out an interior region of the image if requested
		if hasattr(config_module,'mask_radius'):
			x_grid, y_grid = util.make_grid(numPix=image.shape[0],
			deltapix=kwargs_detector['pixel_scale'])
			r = util.array2image(np.sqrt(x_grid**2+y_grid**2))
			image[r<=config_module.mask_radius] = 0

		# Save the image and the metadata
		filename = os.path.join(args.save_folder, 'image_%07d' % nt)
		np.save(filename, image)
		if args.save_png_too:
			plt.imsave(filename + '.png', image)
		for component in sample:
			for key in sample[component]:
				comp_value = sample[component][key]
				# Make sure that lists and other objects that cannot be
				# serialized well are not written out. Warn about this only
				# once.
				if isinstance(comp_value,bool):
					meta_values[component+'_'+key] = int(comp_value)
				elif (isinstance(comp_value,str) or isinstance(comp_value,int) or
					isinstance(comp_value,float)):
					meta_values[component+'_'+key] = comp_value
				elif SERIALIZATIONWARNING:
					warnings.warn('One or more parameters in config_dict '
						'cannot be serialized and will not be written to '
						'metadata.csv',category=RuntimeWarning)
					SERIALIZATIONWARNING = False

		metadata = metadata.append(meta_values,ignore_index=True)

		# Write out the metadata every 20 images
		if nt == 0:
			# Sort the keys lexographically to ensure consistent writes
			metadata = metadata.reindex(sorted(metadata.columns), axis=1)
			metadata.to_csv(metadata_path, index=None)
			metadata = pd.DataFrame()
		elif nt%20 == 0:
			metadata = metadata.reindex(sorted(metadata.columns), axis=1)
			metadata.to_csv(metadata_path, index=None, mode='a',
				header=None)
			metadata = pd.DataFrame()

		nt += 1
		pbar.update()

	# Make sure anything left in the metadata DataFrame is written out
	metadata = metadata.reindex(sorted(metadata.columns), axis=1)
	metadata.to_csv(metadata_path, index=None, mode='a',header=None)
	pbar.close()
	print('Dataset generation complete. Acceptance rate: %.3f'%(args.n/tries))

	# Generate tf record if requested. Save all the parameters and use default
	# filename data.tfrecord
	if args.tf_record:
		# Delayed import, triggers tensorflow import
		from paltas.Analysis import dataset_generation

		# The path to save the TFRecord to.
		tf_record_path = os.path.join(args.save_folder,'data.tfrecord')
		# Generate the list of learning parameters. Only save learning
		# parameters with associated float values.
		learning_params = []
		for key in meta_values:
			if (isinstance(meta_values[key],float) or
				isinstance(meta_values[key],int)):
				learning_params.append(key)
		# Generate the TFRecord
		dataset_generation.generate_tf_record(args.save_folder,learning_params,
			metadata_path,tf_record_path)


if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
"""
Draw from and integrate power laws

Useful equations to draw from and integrate power law distributions.
"""
import numpy as np
import numba


@numba.njit()
def power_law_integrate(p_min,p_max,slope):
	"""Integrates a power law

	Args:
		p_min (float): The lower bound of the power law
		p_max (float): The upper bound of the power law
		slope (float): The slope of the power law

	Returns:
		(float): The integral of the power law x^slope from p_min
		to p_max
	"""
	upper_bound = 1/(slope+1)*p_max**(slope+1)
	lower_bound = 1/(slope+1)*p_min**(slope+1)
	return upper_bound-lower_bound


@numba.njit()
def power_law_draw(p_min,p_max,slope,norm):
	"""Samples from a power law

	Args:
		p_min (float): The lower bound of the power law
		p_max (float): The upper bound of the power law
		slope (float): The slope of the power law
		norm (float): The overall normalization of the power law

	Returns:
		[float,...]: A list of values drawn from the power law
	"""
	# Get the expected number of draws
	n_expected = norm * power_law_integrate(p_min,p_max,slope)

	# Draw the number of objects as a poisson process
	n_draws = np.random.poisson(n_expected)

	# Get the positions in the cdf we want to draw from a uniform and then
	# convert that to values in the pdf
	cdf = np.random.rand(n_draws)
	# usefule shorthand
	s1 = slope + 1
	x = (cdf*(p_max**s1-p_min**s1)+p_min**s1)**(1/s1)

	return x



import glob
import pandas as pd
import os

ss_folders = glob.glob('single_source*')
metadata_list = []

for i,folder in enumerate(ss_folders):
	print(folder)
	# First we want to move all of the numpy images to the desired folder
	npy_files = glob.glob(os.path.join(folder,'*.npy'))
	for j,file in enumerate(npy_files):
		new_path = os.path.join('ss_comp/image_%07d' % (i*50000+j))
		# Don't do this if the file is already there
		if os.path.exists(new_path):
			raise ValueError('%s already exists'%(new_path))
		else:
			print(new_path)
			os.rename(file,new_path)

	# Read the metadata file
	cm = pd.read_csv(os.path.join(folder,'metadata.csv'))
	metadata_list.append(cm)

metadata = pd.concat(metadata_list)
metadata.to_csv('ss_comp/metadata.csv', index=None)


# -*- coding: utf-8 -*-
"""
Utility functions for producing Hubble drizzled images.
"""
import copy
import numpy as np
from astropy.io import fits
from astropy.wcs import wcs
from drizzle.drizzle import Drizzle
from scipy.interpolate import RectBivariateSpline
import warnings
import numba

WCS_ORIGIN = 1


def offset_wcs(w,pixel_offset,reverse=False):
	"""Return wcs w shifted by pixel_offset pixels.

	Args:
		w (astropy.wcs.wcs.WCS): An istance of the WCS class that will
			be copied and offset.
		pixel_offset (tuple): The x,y offset to apply in pixel units.

	Returns:
		(astropy.wcs.wcs.WCS): The offset WCS object.
	"""
	# Copy the object and add the pixel offset to the reference pixel
	# position.
	w_out = copy.deepcopy(w)
	if reverse:
		w_out.wcs.crpix[0] += pixel_offset[1]
		w_out.wcs.crpix[1] += pixel_offset[0]
	else:
		w_out.wcs.crpix += np.array(pixel_offset)
	return w_out


def distort_image(img_high_res,w_high_res,w_dither,offset_pattern,
	psf_supersample_factor):
	"""Create distorted Hubble image in _flt frame

	Args:
		img_high_res (np.array): A high resolution image that will be
			downsampled to produce each of the dithered images.
		w_high_res (astropy.wcs.wcs.WCS): The WCS class for the high resolution
			image. This should not include any distortion effects.
		w_dither (astropy.wcs.wcs.WCS): The WCS class for the dithered image.
			This should have the same pixel scale as the camera and should
			be centered at the same point as w_high_res. The offsets will be
			accounted for in the function call.
		offset_pattern ([tuple,...]): A list of tuples, each containing an x,y
			pair of offsets in pixel coordinates.
		psf_supersample_factor (int): The supersampled resolution at which
			to apply the psf model. If greater than 1 the distorted images
			will be returned at that supersampled scale to allow for
			psf convolution. Nothing needs to be changed about the WCS
			inputs.

	Returns:
		(np.array): A len(offset_pattern) x dither_image.shape sized numpy
		array for each of the dithered images. dither_image shape will
		depend on the w_dither WCS object and the psf_supersample_factor.
	"""
	# First create the interpolator for our image in the sky. Values are assumed
	# to be measured at the center of pixels.
	x_interp = np.arange(img_high_res.shape[0])+0.5
	y_interp = np.arange(img_high_res.shape[1])+0.5
	# Interpolation will be linear.
	img_itp = RectBivariateSpline(x_interp, y_interp, img_high_res,
		kx=1,ky=1)

	# Create the coordinates we want to plot the dithered image on.
	dith_shape = np.array(w_dither.pixel_shape)*psf_supersample_factor
	x_dith,y_dith = np.meshgrid(np.arange(dith_shape[0])+0.5,
		np.arange(dith_shape[1])+0.5,indexing='ij')

	# Array in which we'll store the images
	img_dither_array = np.zeros((len(offset_pattern),)+tuple(dith_shape))

	# Now for each offset requested, generate the dithered image.
	for oi,offset in enumerate(offset_pattern):
		# Calculate the offset WCS
		w_offset = offset_wcs(w_dither,offset)

		# Use that to calculate the ra and dec of each pixel and map
		# that to the image values from the interpolation
		ra_off,dec_off = w_offset.all_pix2world(x_dith/psf_supersample_factor,
			y_dith/psf_supersample_factor,WCS_ORIGIN)
		x_dith_int,y_dith_int = w_high_res.all_world2pix(ra_off,dec_off,
			WCS_ORIGIN)

		img_dither_array[oi] += img_itp(x_dith_int,y_dith_int,grid=False)

	# Multiply by scaling since we want sum not mean.
	img_dither_array *= w_dither.wcs.cdelt[0]/w_high_res.wcs.cdelt[0]
	img_dither_array *= w_dither.wcs.cdelt[1]/w_high_res.wcs.cdelt[1]
	img_dither_array /= psf_supersample_factor**2

	return img_dither_array


def generate_downsampled_wcs(high_res_shape,high_res_pixel_scale,
		low_res_pixel_scale,wcs_distortion):
	"""Generate a wcs mapping onto the specified lower resolution. If
	specified, geometric distortions will be included in the wcs object.

	Args:
		high_res_shape ([int,..]): The shape of the high resolution image
		high_res_pixel_scale (float): The pixel width of the high
			resolution image in units of arcseconds.
		low_res_pixel_scale (float): The pixel width of the lower
			resolution target image in units of arcseconds.
		wcs_distortion (astropy.wcs.wcs.WCS): An instance of the
			WCS class that includes the desired gemoetric distortions as
			SIP coefficients. Note that the parameters relating to
			pixel scale, size of the image, and reference pixels will be
			overwritten. If None no gemoetric distortions will be applied.

	Returns:
		astropy.wcs.wcs.WCS: The wcs object corresponding to the
		downsampled image.
	"""
	# If wcs distortion was specified, use it to seed the fits file we
	# will later use to create the wcs object. This is a little
	# circular, but it's difficult to modify a wcs once it's been
	# created.
	if wcs_distortion is not None:
		warnings.warn('wcs distortion code is not tested.')
		hdul = wcs_distortion.to_fits()
	else:
		hdul = fits.HDUList([fits.PrimaryHDU()])

	# Get the shape of the expected data
	high_res_shape = list(high_res_shape)
	scaling = low_res_pixel_scale/high_res_pixel_scale
	low_res_shape = copy.copy(high_res_shape)
	low_res_shape[0] = low_res_shape[0]//scaling
	low_res_shape[1] = low_res_shape[1]//scaling

	# Modify the fits header to match our desired low resolution
	# configuration.
	hdul[0].header['WCSAXES'] = 2
	hdul[0].header['NAXIS1'] = int(low_res_shape[0])
	hdul[0].header['NAXIS2'] = int(low_res_shape[1])
	hdul[0].header['CRPIX1'] = (low_res_shape[0]/2,
		'Pixel coordinate of reference point')
	hdul[0].header['CRPIX2'] = (low_res_shape[1]/2,
		'Pixel coordinate of reference point')
	hdul[0].header['CDELT1'] = (low_res_pixel_scale/3600,
		'[deg] Coordinate increment at reference point')
	hdul[0].header['CDELT2'] = (low_res_pixel_scale/3600,
		'[deg] Coordinate increment at reference point')
	hdul[0].header['CUNIT1'] = ('deg',
		'Units of coordinate increment and value ')
	hdul[0].header['CUNIT2'] = ('deg',
		'Units of coordinate increment and value ')
	if wcs_distortion:
		hdul[0].header['CTYPE1'] = 'RA-TAN-SIP'
		hdul[0].header['CTYPE2'] = 'DEC-TAN-SIP'
	else:
		hdul[0].header['CTYPE1'] = 'RA'
		hdul[0].header['CTYPE2'] = 'DEC'
	hdul[0].header['CRVAL1'] = (20,
		'[deg] Coordinate value at reference point')
	hdul[0].header['CRVAL2'] = (-70,
		'[deg] Coordinate value at reference point')

	# Use the first object to generate our WCS object. Ignore warnings
	# about not having an actual image in hdul.
	with warnings.catch_warnings():
		warnings.simplefilter('ignore')
		w_ds = wcs.WCS(fobj=hdul,header=hdul[0].header)

	return w_ds


@numba.njit()
def degrade_image(image,degrade_factor):
	"""Degrade an image by the specified integer factor.

	Args:
		image (np.array): The 2D numpy image to degrade
		degrade_factor (int): The integer factor by which
			to degrade the image.

	Returns:
		(np.array): A degraded version of the input image.
	"""
	# Check to make sure the degredation operation is well defined.
	if (image.shape[0] % degrade_factor > 0 or
		image.shape[1] % degrade_factor > 0):
		raise ValueError('Image dimension is not a multiple degrade_factor')

	# Create the new image to save the output to.
	new_image = np.zeros((image.shape[0]//degrade_factor,
			image.shape[1]//degrade_factor))

	# Manually write the image value at each coordinate. For loops are not
	# expensive in numba.
	for i in range(new_image.shape[0]):
		for j in range(new_image.shape[1]):
			new_image[i,j] = np.sum(
				image[degrade_factor*i:degrade_factor*i+degrade_factor,
					degrade_factor*j:degrade_factor*j+degrade_factor])

	return new_image


@numba.njit()
def upsample_image(image,upsample_factor):
	"""Create an upsampled image by repeating the same pixel value
	multiple times.

	Args:
		image (np.array): The 2D numpy image to degrade
		upsample_factor (int): The integer factor by which
			to upsample the image.

	Returns:
		(np.array): A degraded version of the input image.
	"""
	# Create the new image to save the output to.
	new_image = np.zeros((image.shape[0]*upsample_factor,
			image.shape[1]*upsample_factor))

	# Manually write the image value at each coordinate. For loops are not
	# expensive in numba.
	for i in range(new_image.shape[0]):
		for j in range(new_image.shape[1]):
			new_image[i,j] = image[int(i//upsample_factor),
				int(j//upsample_factor)]

	return new_image


def hubblify(img_high_res,high_res_pixel_scale,detector_pixel_scale,
	drizzle_pixel_scale,noise_model,psf_model,offset_pattern,
	wcs_distortion=None,pixfrac=1.0,kernel='square',
	psf_supersample_factor=1):
	"""Generates a simulated drizzled HST image, accounting for
	gemoetric distortions, the dithering pattern, and correlated
	read noise from drizzling.

	Args:
		img_high_res (np.array): A high resolution image that will be
			downsampled to produce each of the dithered images.
		high_res_pixel_scale (float): The pixel width of the
			high resolution image in units of arcseconds.
		detector_pixel_scale (float): The pixel width of the detector
			in units of arcseconds.
		drizzle_pixel_scale (float): The pixel width of the final
			drizzled product in units of arcseconds.
		noise_model (function): A function that maps from an input
			numpy array of the image to a realization of the noise. This
			should operate on the degraded images.
		psf_model (function): A function that maps from an input image
			to an output psf-convovled image. The resolution on which this
			operates is set by psf_supersample_factor. Default is the
			resolution of the detector.
		offset_pattern ([tuple,...]): A list of x,y coordinate pairs
			specifying the offset of each dithered image from the coordinate
			frame used to generate the high resolution image. Specifying shifts
			that place the degraded image outside of the high resolution image
			will cause interpolation errors.
		wcs_distortion (astropy.wcs.wcs.WCS): An instance of the
			WCS class that includes the desired gemoetric distortions.
			Note that the parameters relating to pixel scale, size of
			the image, and reference pixels will be overwritten. If None
			no gemoetric distortions will be applied.
		pixfrac (float): The fraction of the pixel that the pixel
			flux is contained in. Passed to the drizzle algorithm.
		kernel (str): The string for the kernel to be used by the drizzle
			algorithm.
		psf_supersample_factor (int): The supersampled resolution at which
			to apply the psf model. 1 by default (the resolution of the
			detector).

	Returns:
		(np.array): The drizzled image produced from len(offset_pattern)
		number of dithered exposures of img_high_res. Note that this
		means that the drizzled image will have the noise statistics and
		flux of len(offset_pattern) combined exposures.
	"""
	# Create our three base WCS systems, one for the input high res, one for the
	# input highres, one for the drizzled image, and one for the final dithered
	# image.
	w_high_res = generate_downsampled_wcs(img_high_res.shape,
		high_res_pixel_scale,high_res_pixel_scale,None)
	w_driz = generate_downsampled_wcs(img_high_res.shape,high_res_pixel_scale,
		drizzle_pixel_scale,None)
	w_dither = generate_downsampled_wcs(img_high_res.shape,
		high_res_pixel_scale,detector_pixel_scale,wcs_distortion)

	# Initialize our drizzle class with the target output wcs.
	driz = Drizzle(outwcs=w_driz,pixfrac=pixfrac,kernel=kernel,
		fillval='0')

	# Get the distorted sub images to which we will add noise and then add
	# them to our drizzle.
	img_dither_array = distort_image(img_high_res,w_high_res,w_dither,
		offset_pattern,psf_supersample_factor)

	for d_i,image_dither in enumerate(img_dither_array):

		# Add the psf and the noise to each dithered image. Degrade image
		# between psf and noise step if psf is applied on supersampled image.
		image_dither = psf_model(image_dither)
		if psf_supersample_factor > 1:
			image_dither = degrade_image(image_dither,psf_supersample_factor)
		image_dither += noise_model(image_dither)

		# Feed to drizzle, note WCS translation. Also drizzle reverses the
		# axis convention so transpose the image.
		driz.add_image(image_dither.T,offset_wcs(w_dither,
			offset_pattern[d_i]))

	# Get final image from drizzle. Drizzle divides by number of exposures,
	# and we want to undo this effect.
	return driz.outsci.T



from paltas.Configs.xxxx_yyyy.config_val import *

config_dict['subhalo']['parameters']['sigma_sub'] = norm(loc=1.8e-3,
	scale=1.5e-4).rvs
config_dict['main_deflector']['parameters']['gamma'] = truncnorm(-199.5,np.inf,
	loc=1.995,scale=0.01).rvs
config_dict['main_deflector']['parameters']['theta_E'] = truncnorm(-74.9,np.inf,
	loc=1.123,scale=0.015).rvs
config_dict['main_deflector']['parameters']['e1'] = norm(loc=0.031,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['e2'] = norm(loc=-0.003,
	scale=0.01).rvs
config_dict['main_deflector']['parameters']['center_x'] = norm(loc=0.071,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['center_y'] = norm(loc=-0.006,
	scale=0.016).rvs
config_dict['main_deflector']['parameters']['gamma1'] = norm(loc=-0.001,
	scale=0.005).rvs
config_dict['main_deflector']['parameters']['gamma2'] = norm(loc=0.010,
	scale=0.005).rvs



__author__ = 'Sebastian Wagner-Carena'
__email__ = 'sebaswagner@outlook.com'
__version__ = '0.0.3'

# Analysis is not imported by default because it required tensorflow.

from . import Configs
from . import Sampling
from . import Sources
from . import Substructure
from . import Utils
from . import MainDeflector
from . import generate



# manual_keras is intentionally not included. It should be hidden from a high
# level user.
from . import loss_functions
from . import dataset_generation
from . import conv_models
from . import train_model
from . import posterior_functions
from . import hierarchical_inference
from . import pdf_functions



# -*- coding: utf-8 -*-
"""
Turn COSMOS galaxies into Lenstronomy inputs.

This module contains the default class for transforming the objects of the
COSMOS catalog into sources to be passed to lenstronomy.
"""
from pathlib import Path
import astropy
import astropy.table
from astropy.io import fits
from lenstronomy.Util.param_util import phi_q2_ellipticity
import numpy as np
import numpy.lib.recfunctions
from tqdm import tqdm
from .galaxy_catalog import GalaxyCatalog
import scipy

HUBBLE_ACS_PIXEL_WIDTH = 0.03   # Arcsec


class COSMOSCatalog(GalaxyCatalog):
	"""Class to interface to the COSMOS/GREAT3 23.5 magnitude catalog

	This is the catalog used for real galaxies in galsim, see
	https://github.com/GalSim-developers/GalSim/wiki/RealGalaxy%20Data.
	The catalog must be downloaded and unzipped.

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters
			needed to draw sources.

	Notes:

	Required Parameters

	- 	minimum_size_in_pixels - smallest cutout size for COSMOS sources in pixels
	- 	faintest_apparent_mag - faintest apparent AB magnitude for COSMOS source
	- 	max_z - highest redshift for COSMOS source
	- 	smoothing_sigma - smoothing kernel to apply to COSMOS source in pixels
	- 	cosmos_folder - path to COSMOS images
	- 	random_rotation - boolean dictating if COSMOS sources will be rotated
		randomly when drawn
	- 	min_flux_radius - minimum half-light radius for COSMOS source in pixels
	- 	center_x - x-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	center_y- y-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	z_source - source redshift
	"""
	required_parameters = ('minimum_size_in_pixels','faintest_apparent_mag',
		'max_z','smoothing_sigma','cosmos_folder','random_rotation',
		'min_flux_radius','output_ab_zeropoint','z_source','center_x','center_y')
	# Average AB magnitude zeropoint for the COSMOS run.
	ab_zeropoint = 25.95

	def __init__(self, cosmology_parameters, source_parameters):
		super().__init__(cosmology_parameters,source_parameters)

		# Store the path as a Path object.
		self.folder = Path(source_parameters['cosmos_folder'])

		# Check if we've already populated the catalog
		self.catalog_path = self.folder/'paltas_catalog.npy'
		self.npy_files_path = self.folder/'npy_files'
		if self.catalog_path.exists() and self.npy_files_path.exists():
			self.catalog = np.load(str(self.catalog_path))

		else:
			# Make the directory where I'm going to save the npy files
			self.npy_files_path.mkdir()
			# Combine all partial catalog files
			catalogs = [unfits(str(self.folder / fn)) for fn in [
				'real_galaxy_catalog_23.5.fits',
				'real_galaxy_catalog_23.5_fits.fits'
			]]

			# Duplicate IDENT field crashes numpy's silly merge function.
			catalogs[1] = numpy.lib.recfunctions.drop_fields(catalogs[1],
				'IDENT')

			# Custom fields
			catalogs += [
				np.zeros(len(catalogs[0]),
					dtype=[('size_x', np.int),('size_y', np.int),('z', float),
					('pixel_width', float)])]

			self.catalog = numpy.lib.recfunctions.merge_arrays(
				catalogs, flatten=True)

			self.catalog['pixel_width'] = HUBBLE_ACS_PIXEL_WIDTH
			self.catalog['z'] = self.catalog['zphot']

			# Loop over the images to find their sizes.
			catalog_i = 0
			for img, meta in self.iter_image_and_metadata_bulk(
				message='One-time COSMOS startup'):
				# Grab the shape of each image.
				meta['size_x'], meta['size_y'] = img.shape
				# Save the image as its own image.
				img = img.astype(np.float)
				np.save(str(self.npy_files_path/('img_%d.npy'%(catalog_i))),img)
				catalog_i += 1

			np.save(self.catalog_path,self.catalog)

	def __len__(self):
		return len(self.catalog)

	def _passes_cuts(self):
		""" Return a boolean mask of the sources that pass the source
		parameter cuts.

		Returns:
			(np.array): Array of bools of catalog indices to use for
				sampling.
		"""
		is_ok = np.ones(len(self), dtype=np.bool_)
		# Grab the parameter to cut on.
		faintest_apparent_mag = self.source_parameters['faintest_apparent_mag']
		minimum_size_in_pixels = self.source_parameters['minimum_size_in_pixels']
		max_z = self.source_parameters['max_z']
		min_flux_radius = self.source_parameters['min_flux_radius']

		# Get the images that match the cuts.
		if faintest_apparent_mag is not None:
			is_ok &= self.catalog['mag_auto'] < faintest_apparent_mag
		if minimum_size_in_pixels is not None:
			min_size = np.minimum(self.catalog['size_x'],
				self.catalog['size_y'])
			is_ok &= min_size >= minimum_size_in_pixels
		if max_z is not None:
			is_ok &= self.catalog['z'] < max_z
		if min_flux_radius is not None:
			is_ok &= self.catalog['flux_radius'] > min_flux_radius

		return is_ok

	def sample_indices(self,n_galaxies):
		"""Return n_galaxies array of catalog indices to sample

		Args:
			n_galaxies (int): Number of indices to return

		Returns:
			(np.array): Array of ints of catalog indices to sample.

		Notes:
			The minimum apparent magnitude, minimum size in pixels, and
			minimum redshift are all set by the source parameters dict.
		"""
		is_ok = self._passes_cuts()
		return np.random.choice(np.where(is_ok)[0],size=n_galaxies,
			replace=True)

	@staticmethod
	def _file_number(fn):
		"""Return integer X in blah_nX.fits filename fn.
		X can be more than one digit, not necessarily zero padded.
		"""
		return int(str(fn).split('_n')[-1].split('.')[0])

	def iter_image_and_metadata_bulk(self, message=''):
		"""Yields the image array and metadata for all of the images
		in the catalog.

		Args:
			message (str): If the iterator uses tqdm, this message
				will be displayed.

		Returns:
			(generator): A generator that can be iterated over to give
			lenstronomy kwargs.

		Notes:
			This will read the fits files.
		"""
		catalog_i = 0
		_pattern = f'real_galaxy_images_23.5_n*.fits'  # noqa: F541, F999
		files = list(sorted(self.folder.glob(_pattern),
			key=self._file_number))

		# Iterate over all the matching files.
		for fn in tqdm(files, desc=message):
			with fits.open(fn) as hdul:
				for img in hdul:
					yield img.data, self.catalog[catalog_i]
					catalog_i += 1

	def image_and_metadata(self, catalog_i):
		"""Returns the image array and metadata for one galaxy

		Args:
			catalog_i (int): The catalog index

		Returns:
			([np.array, np.void]) A numpy array containing the image
			metadata and a numpy void type that acts as a dictionary with
			the metadata.

		Notes:
			This will read the numpy files made during initialization. This is
			much faster on average than going for the fits files.
		"""
		img = np.load(str(self.npy_files_path/('img_%d.npy'%(catalog_i))))
		smoothing_sigma = self.source_parameters['smoothing_sigma']
		# Filter the images if that was requested
		if smoothing_sigma > 0:
			img = scipy.ndimage.gaussian_filter(img,
				sigma=smoothing_sigma/HUBBLE_ACS_PIXEL_WIDTH)
		return img, self.catalog[catalog_i]


class COSMOSSersicCatalog(COSMOSCatalog):
	"""Class identical to COSMOSCatalog, but produces the best-fit single
	elliptic Sersic profiles instead of real galaxy images

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters
			needed to draw sources.

	Notes:

	Required Parameters

	- 	minimum_size_in_pixels - smallest cutout size for COSMOS sources in pixels
	- 	faintest_apparent_mag - faintest apparent AB magnitude for COSMOS source
	- 	max_z - highest redshift for COSMOS source
	- 	smoothing_sigma - smoothing kernel to apply to COSMOS source in pixels
	- 	cosmos_folder - path to COSMOS images
	- 	random_rotation - boolean dictating if COSMOS sources will be rotated
		randomly when drawn
	- 	min_flux_radius - minimum half-light radius for COSMOS source in pixels
	- 	center_x - x-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	center_y- y-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	z_source - source redshift
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		_fit_results = self.catalog['sersicfit'].astype(np.float)
		self.sercic_info = {
			p: _fit_results[:, i]
			for i, p in enumerate(
				'intensity r_half n q boxiness x0 y0 phi'.split())}
		# Convert half-light radius from pixels to arcseconds
		self.sercic_info['r_half'] *= HUBBLE_ACS_PIXEL_WIDTH

	def draw_source(self, catalog_i=None, phi=None):
		"""Creates lenstronomy interpolation lightmodel kwargs from
		a catalog image.

		Args:
			catalog_i (int): Index of image in catalog
			z_new (float): Redshift to place image at
			phi (float): Rotation to apply to the image.
				If not provided, use random or original rotation
				depending on source_parameters['random_rotation']

		Returns:
			(list,list) A list containing the model ['INTERPOL'] and
			the kwargs for an instance of the class
			lenstronomy.LightModel.Profiles.interpolation.Interpol

		Notes:
			If not catalog_i is provided, one that meets the cuts will be
			selected at random.
		"""
		catalog_i, phi = self.fill_catalog_i_phi_defaults(catalog_i, phi)
		metadata = self.catalog[catalog_i]
		z_new = self.source_parameters['z_source']

		z_scaling = self.z_scale_factor(metadata['z'], z_new)

		# Get sercic info for this particular galaxy
		sercic_info = {
			p: self.sercic_info[p][catalog_i]
			for p in self.sercic_info}

		# Convert (phi, q) -> (e1, e2), after applying possible random rotation
		# Using py_func to avoid numba caching trouble
		# (and it's a trivial func anyway)
		e1, e2 = phi_q2_ellipticity.py_func(
			(sercic_info['phi'] + phi) % (2 * np.pi),
			sercic_info['q'])

		# Convert to kwargs for lenstronomy
		return (
			['SERSIC_ELLIPSE'],
			[dict(
				# Scale by pixel area before z-scaling, as in parent draw_source
				amp=sercic_info['intensity'] / metadata['pixel_width'] ** 2,
				e1=e1,
				e2=e2,
				R_sersic=sercic_info['r_half'] * z_scaling,
				n_sersic=sercic_info['n'])])

	def image_and_metadata(self, catalog_i):
		"""Returns the image array and metadata for one galaxy

		Args:
			catalog_i (int): The catalog index

		Returns:
			([np.array, np.void]) A numpy array containing the image
			metadata and a numpy void type that acts as a dictionary with
			the metadata.

		Notes:
			This will read the numpy files made during initialization. This is
			much faster on average than going for the fits files.
		"""
		raise NotImplementedError

	def iter_image_and_metadata_bulk(self, message=''):
		"""Yields the image array and metadata for all of the images
		in the catalog.

		Args:
			message (str): If the iterator uses tqdm, this message
				will be displayed.

		Returns:
			(generator): A generator that can be iterated over to give
			lenstronomy kwargs.

		Notes:
			This will read the fits files.
		"""
		raise NotImplementedError


class COSMOSExcludeCatalog(COSMOSCatalog):
	"""	Class identical to COSMOSCatalog, but allows for specific source indexes
	to be excluded from the analysis.

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters
			needed to draw sources.

	Notes:

	Required Parameters

	- 	minimum_size_in_pixels - smallest cutout size for COSMOS sources in pixels
	- 	faintest_apparent_mag - faintest apparent AB magnitude for COSMOS source
	- 	max_z - highest redshift for COSMOS source
	- 	smoothing_sigma - smoothing kernel to apply to COSMOS source in pixels
	- 	cosmos_folder - path to COSMOS images
	- 	random_rotation - boolean dictating if COSMOS sources will be rotated
		randomly when drawn
	- 	min_flux_radius - minimum half-light radius for COSMOS source in pixels
	- 	center_x - x-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	center_y- y-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	z_source - source redshift
	-	source_exclusion_list - A list of COSMOS source indices to exclude,
		even if they pass the other cuts.
	"""

	required_parameters = ('minimum_size_in_pixels','faintest_apparent_mag',
		'max_z','smoothing_sigma','cosmos_folder','random_rotation',
		'min_flux_radius','source_exclusion_list','output_ab_zeropoint',
		'center_x','center_y','z_source')

	def __init__(self, cosmology_parameters, source_parameters):
		super().__init__(cosmology_parameters,source_parameters)

	def _passes_cuts(self):
		""" Return a boolean mask of the sources that pass the source
		parameter cuts.

		Returns:
			(np.array): Array of bools of catalog indices to use for
				sampling.
		"""
		is_ok = super()._passes_cuts()

		# Drop any catalog indices that are in the exclusion list
		is_ok &= np.invert(np.in1d(np.arange(len(self)),
			self.source_parameters['source_exclusion_list']))

		return is_ok


class COSMOSIncludeCatalog(COSMOSCatalog):
	"""Class identical to COSMOSExcludeCatalog, but only allows for specific
	source indeces to be included for the analysis.

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters
			needed to draw sources.

		Notes:

	Required Parameters

	- 	minimum_size_in_pixels - smallest cutout size for COSMOS sources in pixels
	- 	faintest_apparent_mag - faintest apparent AB magnitude for COSMOS source
	- 	max_z - highest redshift for COSMOS source
	- 	smoothing_sigma - smoothing kernel to apply to COSMOS source in pixels
	- 	cosmos_folder - path to COSMOS images
	- 	random_rotation - boolean dictating if COSMOS sources will be rotated
		randomly when drawn
	- 	min_flux_radius - minimum half-light radius for COSMOS source in pixels
	- 	center_x - x-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	center_y- y-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	z_source - source redshift
	-	source_inclusion_list - A list of COSMOS source indices to include,
		cutting any source not in this list.
	"""

	required_parameters = ('minimum_size_in_pixels','faintest_apparent_mag',
		'max_z','smoothing_sigma','cosmos_folder','random_rotation',
		'min_flux_radius','source_inclusion_list','output_ab_zeropoint',
		'center_x','center_y','z_source')

	def __init__(self, cosmology_parameters, source_parameters):
		super().__init__(cosmology_parameters,source_parameters)

	def _passes_cuts(self):
		""" Return a boolean mask of the sources that pass the source
		parameter cuts.

		Returns:
			(np.array): Array of bools of catalog indices to use for
				sampling.
		"""
		is_ok = super()._passes_cuts()

		# Drop any catalog indices that are in the exclusion list
		is_ok &= np.in1d(np.arange(len(self)),
			self.source_parameters['source_inclusion_list'])

		return is_ok


def unfits(fn, pandas=False):
	"""Returns numpy record array from fits catalog file fn

	Args:
		fn (str): filename of fits file to load
		pandas (bool): If True, return pandas DataFrame instead of an array

	Returns:
		(np.array):
	"""
	if pandas:
		astropy.table.Table.read(fn, format='fits').to_pandas()
	else:
		with fits.open(fn) as hdul:
			data = hdul[1].data
			# Remove fitsyness from record array
			return np.array(data, dtype=data.dtype)



# Includes a PEMD deflector with external shear, and Sersic sources. Includes 
# a simple observational effect model that roughly matches HST effects for
# Wide Field Camera 3 (WFC3) IR channel with the F160W filter.

import numpy as np
from scipy.stats import norm, truncnorm, uniform
from paltas.MainDeflector.simple_deflectors import PEMDShear
from paltas.Sources.sersic import SingleSersicSource

# Define the numerics kwargs.
kwargs_numerics = {'supersampling_factor':1}

# This is always the number of pixels for the CCD. If drizzle is used, the
# final image will be larger.
numpix = 128

# Define some general image kwargs for the dataset
mask_radius = 0.5
mag_cut = 2.0

# Define arguments that will be used multiple times
output_ab_zeropoint = 25.127

config_dict = {
	'main_deflector':{
		'class': PEMDShear,
		'parameters':{
			'M200': 1e13,
			'z_lens': 0.5,
			'gamma': truncnorm(-20,np.inf,loc=2.0,scale=0.1).rvs,
			'theta_E': truncnorm(-1.1/0.15,np.inf,loc=1.1,scale=0.15).rvs,
			'e1': norm(loc=0.0,scale=0.1).rvs,
			'e2': norm(loc=0.0,scale=0.1).rvs,
			'center_x': norm(loc=0.0,scale=0.16).rvs,
			'center_y': norm(loc=0.0,scale=0.16).rvs,
			'gamma1': norm(loc=0.0,scale=0.05).rvs,
			'gamma2': norm(loc=0.0,scale=0.05).rvs,
			'ra_0':0.0, 'dec_0':0.0
		}
	},
	'source':{
		'class': SingleSersicSource,
		'parameters':{
			'z_source':truncnorm(-5,np.inf,loc=2.,scale=0.4).rvs,
			'magnitude':uniform(loc=20,scale=5).rvs,
			'output_ab_zeropoint':output_ab_zeropoint,
			'R_sersic':truncnorm(-2,2,loc=0.35,scale=0.05).rvs,
			'n_sersic':truncnorm(-6.,np.inf,loc=3.,scale=0.5).rvs,
			'e1':norm(loc=0.0,scale=0.1).rvs,
			'e2':norm(loc=0.0,scale=0.1).rvs,
			'center_x':norm(loc=0.0,scale=0.16).rvs,
			'center_y':norm(loc=0.0,scale=0.16).rvs}
	},
	'cosmology':{
		'parameters':{
			'cosmology_name': 'planck18'
		}
	},
	'psf':{
		'parameters':{
			'psf_type':'GAUSSIAN',
			'fwhm': 0.03
		}
	},
	'detector':{
		'parameters':{
			'pixel_scale':0.040,'ccd_gain':1.58,'read_noise':3.0,
			'magnitude_zero_point':output_ab_zeropoint,
			'exposure_time':1380,'sky_brightness':21.83,
			'num_exposures':4,'background_noise':None
		}
	}
}



import numpy as np
import os, argparse, sys
import numba
from manada.Analysis import hierarchical_inference
import emcee
from multiprocessing import Pool
from importlib import import_module


def parse_args():
	"""Parse the input arguments by the user

	Returns:
		(argparse.Namespace): An instance of the Namespace object with the
			users provided values.

	"""
	# Initialize the parser and the possible inputs
	parser = argparse.ArgumentParser()
	parser.add_argument('test_index',help='The test set to run hierarchical ' +
		'inference on.',type=int)
	parser.add_argument('--n_samps', default=10000, type=int, dest='n_samps',
		help='Number of emcee samples to draw.')
	parser.add_argument('--n_lenses', default=10, type=int,dest='n_lenses',
		help='Number of lenses to include in analysis.')
	args = parser.parse_args()
	return args


def main():
	# Get the arguments
	args = parse_args()

	# Extract the parameters from the input args.
	n_lenses = args.n_lenses
	n_emcee_samps = args.n_samps
	chains_folder = '/scratch/users/swagnerc/manada/chains'
	test_index = args.test_index
	backend_path = 'test_set_%d_lenses_%d.h5'%(test_index,n_lenses)
	chains_path = os.path.join(chains_folder,backend_path)
	n_ensemble = 5

	# Load the predictions for the mean and covariance for one of our
	# ensemble models.
	mi = 3
	y_pred = (np.load('/scratch/users/swagnerc/manada/chains/'
		+ 'marg_outputs/y_pred_marg_model_%d_set_%d.npy'%(mi,
			test_index)).astype(np.float64))
	cov_pred = (np.load('/scratch/users/swagnerc/manada/chains/' +
		'marg_outputs/y_cov_marg_model_%d_set_%d.npy'%(mi,
			test_index)).astype(np.float64))
	prec_pred = (np.linalg.inv(cov_pred))

	# Pull the target parameters straight from the config file.
	sys.path.insert(0,'/scratch/users/swagnerc/manada/datasets/marg_ns/' +
		'marg_ns_test_shift_%d/'%(test_index))
	config_module = import_module('config_marg_shift_%d'%(test_index))
	md_params = config_module.config_dict['main_deflector']['parameters']
	subhalo_params = config_module.config_dict['subhalo']['parameters']

	# The interim training distribution.
	mu_omega_i = np.array([1.1,0.0,0.0,2.0,0.0,0.0,0.0,0.0,2e-3])
	cov_omega_i =np.diag(np.array([0.15,0.05,0.05,0.1,0.1,0.1,0.16,0.16,
		1.1e-3])**2)

	# We will want to initialize emcee near the correct values.
	mu_omega = np.array([md_params['theta_E'].__self__.mean(),
		md_params['gamma1'].__self__.mean(),md_params['gamma2'].__self__.mean(),
		md_params['gamma'].__self__.mean(),md_params['e1'].__self__.mean(),
		md_params['e2'].__self__.mean(),md_params['center_x'].__self__.mean(),
		md_params['center_y'].__self__.mean(),
		subhalo_params['sigma_sub'].__self__.mean()])
	cov_omega =np.diag(np.array([md_params['theta_E'].__self__.std(),
		md_params['gamma1'].__self__.std(),md_params['gamma2'].__self__.std(),
		md_params['gamma'].__self__.std(),md_params['e1'].__self__.std(),
		md_params['e2'].__self__.std(),md_params['center_x'].__self__.std(),
		md_params['center_y'].__self__.std(),
		subhalo_params['sigma_sub'].__self__.std()])**2)
	true_hyperparameters = np.concatenate([mu_omega,
		np.log(np.diag(np.sqrt(cov_omega)))])

	# A prior function that mainly just bounds the uncertainty estimation.
	@numba.njit()
	def eval_func_omega(hyperparameters):
		upper_lim = np.log(np.diag(np.sqrt(cov_omega_i)))+0.5
		lower_lim = np.log(np.diag(np.sqrt(cov_omega)))-0.2
		if (np.any(hyperparameters[len(hyperparameters)//2:]>upper_lim) or
			np.any(hyperparameters[len(hyperparameters)//2:]<lower_lim)):
			return -np.inf
		# Make sure the means don't drift too far into absurdity
		if (np.any(hyperparameters[:len(hyperparameters)//2]>mu_omega_i+
				np.sqrt(np.diag(cov_omega_i))*5) or
			np.any(hyperparameters[:len(hyperparameters)//2]<mu_omega_i-
				np.sqrt(np.diag(cov_omega_i))*5)):
			return -np.inf
		return 0

	prob_class = hierarchical_inference.ProbabilityClassAnalytical(mu_omega_i,
		cov_omega_i,eval_func_omega)
	prob_class.set_predictions(mu_pred_array_input=y_pred[:,:n_lenses],
		prec_pred_array_input=prec_pred[:,:n_lenses])

	with Pool() as pool:
		n_walkers = 40
		ndim = 18
		initial_std = true_hyperparameters*0.002
		initial_std[initial_std==0] += 0.002
		initial_std[:len(initial_std)//2] *= 5
		cur_state = (np.random.rand(n_walkers, ndim)*2-1)*initial_std
		cur_state += true_hyperparameters
		backend = emcee.backends.HDFBackend(chains_path)
		sampler = emcee.EnsembleSampler(n_walkers, ndim,prob_class.log_post_omega,
			backend=backend,pool=pool)
		sampler.run_mcmc(cur_state,n_emcee_samps,progress=True,
			skip_initial_state_check=True)


if __name__ == '__main__':
	main()



# -*- coding: utf-8 -*-
"""
Generate Lenstronomy inputs for a COSMOS galaxy + a Sersic Ellipse

This module contains a class that combines a COSMOS galaxy and a Sersic light 
distribution as the two sources for paltas.
"""
from .cosmos import COSMOSCatalog
from .sersic import SingleSersicSource


class COSMOSSersic(COSMOSCatalog):
	"""Class to Combine COSMOS galaxy with Sersic light source

	Args:
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology): 
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
		source_parameters (dict): A dictionary containing all the parameters 
			needed to draw sources.

	Notes:

	Required Parameters

	- 	minimum_size_in_pixels - smallest cutout size for COSMOS sources in pixels
	- 	faintest_apparent_mag - faintest apparent AB magnitude for COSMOS source
	- 	max_z - highest redshift for COSMOS source
	- 	smoothing_sigma - smoothing kernel to apply to COSMOS source in pixels
	- 	cosmos_folder - path to COSMOS images
	- 	random_rotation - boolean dictating if COSMOS sources will be rotated
		randomly when drawn
	- 	min_flux_radius - minimum half-light radius for COSMOS source in pixels
	- 	center_x - x-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	center_y- y-coordinate lens center for COSMOS source in units of
		arcseconds
	- 	mag_sersic - AB absolute magnitude of the sersic
	- 	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	R_sersic - Sersic radius in units of arcseconds
	- 	n_sersic - Sersic index
	- 	e1_sersic - x-direction ellipticity eccentricity for sersic
	- 	e2_sersic - xy-direction ellipticity eccentricity for sersic
	- 	center_x_sersic - x-coordinate lens center for sersic in units of
		arcseconds
	- 	center_y_sersic - y-coordinate lens center for sersic in units of
		arcseconds
	- 	z_source - source redshift
	"""

	required_parameters = ('minimum_size_in_pixels','faintest_apparent_mag',
		'max_z','smoothing_sigma','cosmos_folder','random_rotation',
		'min_flux_radius','output_ab_zeropoint','z_source', 'mag_sersic',
		'center_x','center_y','R_sersic', 'n_sersic','e1_sersic', 'e2_sersic',
		'center_x_sersic','center_y_sersic')

	def draw_source(self, catalog_i=None, phi=None):
		"""Creates lenstronomy kwargs for a COSMOS catalog image and a 
		Sersic source

		Args:
			catalog_i (int): Index of image in catalog
			phi (float): Rotation to apply to the image.
				If not provided, use random or original rotation
				depending on source_parameters['random_rotation']

		Returns:
			(list,list) A list containing the models ['INTERPOL',
			'SERSIC_ELLIPSE'] and a list of two kwarg dicts for the
			instances of the lenstronomy classes

		Notes:
			If catalog_i is not provided, one that meets the cuts will be
			selected at random.
		"""
		model_list, kwargs_list = COSMOSCatalog.draw_source(self, catalog_i, phi)
		# look up difference between append & extend
		model_list.append('SERSIC_ELLIPSE')

		# create dict for sersic
		sersic_kwargs_dict = {}
		suffix = '_sersic'
		for param_name in self.__class__.required_parameters:
			if suffix in param_name:
				# R_sersic & n_sersic keep suffix
				if (param_name == 'R_sersic' or param_name == 'n_sersic'):
					sersic_kwargs_dict[param_name] = (
						self.source_parameters[param_name])
				else:
					sersic_kwargs_dict[param_name[:-7]] = (
						self.source_parameters[param_name])

		# mag to amp conversion
		sersic_kwargs_dict.pop('mag')
		sersic_kwargs_dict['amp'] = SingleSersicSource.mag_to_amplitude(
			self.source_parameters['mag_sersic'],
			self.source_parameters['output_ab_zeropoint'], sersic_kwargs_dict)

		kwargs_list.append(sersic_kwargs_dict)

		return model_list, kwargs_list



# -*- coding: utf-8 -*-
"""
Manipulate the posterios produced by the network to test for calibration.

This module contains functions to test the calibration of the outputs of the
network on a test or validation set.
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import linregress
import numba


def plot_coverage(y_pred,y_test,std_pred,parameter_names,
	color_map=["#377eb8", "#4daf4a","#e41a1c","#984ea3"],block=True,
	fontsize=20,show_error_bars=True,n_rows=4):
	""" Generate plots for the 1D coverage of each parameter.

	Args:
		y_pred (np.array): A (batch_size,num_params) array containing the
			mean prediction for each Gaussian
		y_test (np.array): A (batch_size,num_params) array containing the
			true value of the parameter on the test set.
		std_pred (np.array): A (batch_size,num_params) array containing the
			predicted standard deviation for each parameter.
		parameter_names ([str,...]): A list of the parameter names to be
			printed in the plots.ed.
		color_map ([str,...]): A list of at least 4 colors that will be used
			for plotting the different coverage probabilities.
		block (bool): If true, block excecution after plt.show() command.
		fontsize (int): The fontsize to use for the parameter names.
		show_error_bars (bool): If true plot the error bars on the coverage
			plot.
		n_rows (int): The number of rows to include in the subplot.
	"""
	num_params = len(parameter_names)
	error = y_pred - y_test
	# Define the covariance masks for our coverage plots.
	cov_masks = [np.abs(error)<=std_pred,np.abs(error)<2*std_pred,
		np.abs(error)<3*std_pred, np.abs(error)>=3*std_pred]
	cov_masks_names = ['1 sigma =', '2 sigma =', '3 sigma =', '>3 sigma =']
	for i in range(len(parameter_names)):
		plt.subplot(n_rows, int(np.ceil(num_params/n_rows)), i+1)
		# Plot the datapoints in each coverage interval seperately.
		for cm_i in range(len(cov_masks)-1,-1,-1):
			cov_mask = cov_masks[cm_i][:,i]
			yt_plot = y_test[cov_mask,i]
			yp_plot = y_pred[cov_mask,i]
			ys_plot = std_pred[cov_mask,i]
			# Plot with errorbars if requested
			if show_error_bars:
				plt.errorbar(yt_plot,yp_plot,yerr=ys_plot, fmt='.',
					c=color_map[cm_i],
					label=cov_masks_names[cm_i]+'%.2f'%(
						np.sum(cov_mask)/len(error)))
			else:
				plt.errorbar(yt_plot,yp_plot,fmt='.',c=color_map[cm_i],
					label=cov_masks_names[cm_i]+'%.2f'%(
						np.sum(cov_mask)/len(error)))

		# Include the correlation coefficient squared value in the plot.
		_, _, rval, _, _ = linregress(y_test[:,i],y_pred[:,i])
		straight = np.linspace(np.min(y_test[:,i]),np.max(y_test[:,i]),10)
		plt.plot(straight, straight, label='',color='k')
		plt.text(0.8*np.max(straight)+0.2*np.min(straight),np.min(straight),
			'$R^2$: %.3f'%(rval**2),{'fontsize':fontsize})
		plt.title(parameter_names[i],fontsize=fontsize)
		plt.ylabel('Prediction',fontsize=fontsize)
		plt.xlabel('True Value',fontsize=fontsize)
		plt.legend(**{'fontsize':fontsize},loc=2)
	plt.show(block=block)


def calc_p_dlt(predict_samps,y_test,weights=None,cov_dist_mat=None):
	"""	Calculate the percentage of draws whose radial distance
	(weighted by the covariance) is larger than that of the truth.

	Args:
		predict_samps (np.array): An array with dimensions (n_samples,
			batch_size,num_params) containing samples drawn from the
			predicted distribution.
		y_test (np.array): A (batch_size,num_params) array containing the
			true value of the parameter on the test set.
		weights (np.array): An array with dimension (n_samples,
			batch_size) that will be used to reweight the posterior.
		cov_dist_mat (np.array): A (num_params,num_params) array representing
			the covariance matrix to use for the distance calculation. If None
			a covariance matrix will be estimated from the y_test data.

	Returns:
		(np.array): A array of length batch_size containing the percentage
		of draws with larger radial distance.
	"""
	# Factor weights into the mean.
	if weights is None:
		y_mean = np.mean(predict_samps,axis=0)
	else:
		y_mean = np.mean(np.expand_dims(weights,axis=-1)*predict_samps,axis=0)

	# The metric for the distance calculation. Using numba for speed.
	@numba.njit
	def d_m(dif,cov):
		d_metric = np.zeros(dif.shape[0:2])
		for i in range(d_metric.shape[0]):
			for j in range(d_metric.shape[1]):
				d_metric[i,j] = np.dot(dif[i,j],np.dot(cov,dif[i,j]))
		return d_metric

	# Use emperical covariance for distance metric unless matrix was passed
	# in.
	if cov_dist_mat is None:
		cov_emp = np.cov(y_test.T)
	else:
		cov_emp = cov_dist_mat

	p_dlt = (d_m(predict_samps-y_mean,cov_emp)<
		d_m(np.expand_dims(y_test-y_mean,axis=0),cov_emp))

	# Calculate p_dlt factoring in the weights if needed.
	if weights is None:
		return np.mean(p_dlt,axis=0)
	else:
		return np.mean(p_dlt*weights,axis=0)


def plot_calibration(predict_samps,y_test,color_map=["#377eb8", "#4daf4a"],
	n_perc_points=20,figure=None,legend=None,show_plot=True,block=True,
	weights=None,title=None,ls='-',loc=9,dpi=200):
	"""	Plot the multidimensional calibration of the neural network predicted
	posteriors.

	Args:
		predict_samps (np.array): An array with dimensions (n_samples,
			batch_size,num_params) containing samples drawn from the
			predicted distribution.
		y_test (np.array): A (batch_size,num_params) array containing the
			true value of the parameter on the test set.
		color_map ([str,...]): A list of the colors to use in plotting.
		n_perc_point (int): The number of percentages to probe in the
			plotting.
		figure (matplotlib.pyplot.figure): A figure that was previously
			returned by plot_calibration to overplot onto.
		legend ([str,...]): The legend to use for plotting.
		show_plot (bool): If true, call plt.show() at the end of the
			function.
		block (bool): If true, block excecution after plt.show() command.
		weights (np.array): An array with dimension (n_samples,
			batch_size) that will be used to reweight the posterior.
		title (str): The title to use for the plot. If None will use a
			default title.
		ls (str): The line style to use in the calibration line for the
			BNN.
		loc (int or tuple): The location for the legend in the calibration
			plot.
		dpi (int): The dpi to use for the figure.

	Returns:
		(np.array): A array of length batch_size containing the percentage
		of draws with larger radial distance.
	"""
	# Go through each of our examples and see what percentage of draws have
	# ||draws||_2 < ||truth||_2 (essentially doing integration by radius).
	p_dlt = calc_p_dlt(predict_samps,y_test,weights)

	# Plot what percentage of images have at most x% of draws with
	# p(draws)>p(true).
	percentages = np.linspace(0.0,1.0,n_perc_points)
	p_images = np.zeros_like(percentages)
	if figure is None:
		fig = plt.figure(figsize=(8,8),dpi=dpi)
		plt.plot(percentages,percentages,c=color_map[0],ls='--')
	else:
		fig = figure

	# We'll estimate the uncertainty in our plat using a jacknife method.
	p_images_jn = np.zeros((len(p_dlt),n_perc_points))
	for pi in range(n_perc_points):
		percent = percentages[pi]
		p_images[pi] = np.mean(p_dlt<=percent)
		for ji in range(len(p_dlt)):
			samp_p_dlt = np.delete(p_dlt,ji)
			p_images_jn[ji,pi] = np.mean(samp_p_dlt<=percent)

	# Estimate the standard deviation from the jacknife
	p_dlt_std = np.sqrt((len(p_dlt)-1)*np.mean(np.square(p_images_jn-
		np.mean(p_images_jn,axis=0)),axis=0))
	plt.plot(percentages,p_images,c=color_map[1],ls=ls)

	# Plot the 1 sigma contours from the jacknife estimate to get an idea of
	# our sample variance.
	plt.fill_between(percentages,p_images+p_dlt_std,p_images-p_dlt_std,
		color=color_map[1],alpha=0.3)

	if figure is None:
		plt.xlabel('Percentage of Probability Volume')
		plt.ylabel('Percent of Lenses With True Value in the Volume')
		plt.text(-0.03,1,'Underconfident')
		plt.text(0.80,0,'Overconfident')
	if title is None:
		plt.title('Calibration of Network Posterior')
	else:
		plt.title(title)
	if legend is None:
		plt.legend(['Perfect Calibration','Network Calibration'],
			loc=loc)
	else:
		plt.legend(legend,loc=loc)
	if show_plot:
		plt.show(block=block)

	return fig



from paltas.Configs.xxxx_yyyy.config_train import *
from paltas.Sources.cosmos import COSMOSIncludeCatalog
import os
import pandas as pd

# config_dict and root path in the import *.
config_dict['source']['parameters']['source_inclusion_list'] = pd.read_csv(
	os.path.join(root_path,'paltas/Sources/val_galaxies.csv'),
	names=['catalog_i'])['catalog_i'].to_numpy()
config_dict['source']['class'] = COSMOSIncludeCatalog
del config_dict['source']['parameters']['source_exclusion_list']



# -*- coding: utf-8 -*-
"""
Provide the base class for specifying the main deflector of a lensing system.

This module contains the base class that all the main deflector classes will
build from. Because the steps for rendering a main deflector can vary between
different models, the required functions are very sparse.
"""
from ..Utils.cosmology_utils import get_cosmology
import copy


class MainDeflectorBase():
	"""Base class for rendering the main halo.

	Args:
		main_deflector_parameters (dict): A dictionary containing the type of
			los distribution and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	"""

	required_parameters = tuple()

	def __init__(self,main_deflector_parameters,cosmology_parameters):

		# Save the parameters as a copy to avoid any misunderstanding on the
		# user end.
		self.main_deflector_parameters = copy.deepcopy(
			main_deflector_parameters)

		# Turn our cosmology parameters into a colossus cosmology instance
		self.cosmo = get_cosmology(cosmology_parameters)

		# Check that all the needed parameters are present
		self.check_parameterization(self.__class__.required_parameters)

	def check_parameterization(self,required_params):
		""" Check that all the required parameters are present in the
		los_parameters.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		if not all(elem in self.main_deflector_parameters.keys() for
			elem in required_params):
			raise ValueError('Not all of the required parameters for the ' +
				'parameterization are present.')

	def update_parameters(self,main_deflector_parameters=None,
		cosmology_parameters=None):
		"""Updated the class parameters

		Args:
			los_parameters (dict): A dictionary containing the type of
				los distribution and the value for each of its parameters.
			main_deflector_parameters (dict): A dictionary containing the type
				of main deflector and the value for each of its parameters.
			source_parameters (dict): A dictionary containing the type of the
				source and the value for each of its parameters.
			cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
				Either a name of colossus cosmology, a dict with 'cosmology name':
				name of colossus cosmology, an instance of colussus cosmology, or
				a dict with H0 and Om0 ( other parameters will be set to
				defaults).
		"""
		if main_deflector_parameters is not None:
			self.main_deflector_parameters = copy.copy(
				main_deflector_parameters)
		if cosmology_parameters is not None:
			self.cosmo = get_cosmology(cosmology_parameters)

	def draw_main_deflector(self):
		"""Draws the lenstronomy profile names and kwargs for the components
		of the main deflector.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each component of the main deflector, the second is the
			lenstronomy kwargs for each component, and the third is the
			redshift for each component.
		"""
		raise NotImplementedError



# -*- coding: utf-8 -*-
"""
Implement a few models for parameter inference.

This module implements models to be used for analysis of strong
lensing parameters.
"""

from tensorflow.keras import layers
from tensorflow.keras.models import Model
import tensorflow as tf


def _xresnet_block(x,filters,kernel_size,strides,conv_shortcut,name,
	trainable=True):
	""" Build a block of residual convolutions for the xresnet model family

	Args:
		x (KerasTensor): The input Keras tensorflow tensor that will be
			passed through the stack.
		filters (int): The number of output filters
		kernel_size (int): The kernel size of each filter
		strides (int): The strides to use for each convolutional filter
		conv_shortcut (bool): When true, will use a convolutional shortcut and
			when false will use an identity shortcut.
		name (str): The name for this block
		trainable (bool): If false, the weights in this block will not be
			trainable.

	Returns:
		(KerasTensor): A Keras tensorflow tensor representing the input after
		the block has been applied
	"""
	# First axis is assumed to be the batch
	bn_axis = -1

	# Use the ResnetD variant for the shortcut
	shortcut = x
	if strides > 1:
		shortcut = layers.AveragePooling2D(pool_size=(2,2),name=name+'_id_pool',
			trainable=trainable)(shortcut)
	if conv_shortcut is True:
		shortcut = layers.Conv2D(filters,1,strides=1,use_bias=False,
			name=name+'_id_conv',trainable=trainable)(shortcut)
		shortcut = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,
			momentum=0.1,name=name+'_id_bn',trainable=trainable)(shortcut)

	# Set up the rest of the block
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name=name+'_pad1',
		trainable=trainable)(x)
	x = layers.Conv2D(filters,kernel_size,strides=strides,use_bias=False,
		name=name+'_conv1',trainable=trainable)(x)
	x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
		name=name+'_bn1',trainable=trainable)(x)
	x = layers.Activation('relu',name=name+'_relu1',trainable=trainable)(x)
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name=name+'_pad2',
		trainable=trainable)(x)
	x = layers.Conv2D(filters,kernel_size,strides=1,use_bias=False,
		name=name+'_conv2',trainable=trainable)(x)
	x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
		name=name+'_bn2',gamma_initializer='zeros',trainable=trainable)(x)

	x = layers.Add(name=name+'_add',trainable=trainable)([shortcut,x])
	x = layers.Activation('relu',name=name+'_out',trainable=trainable)(x)

	return x


def _xresnet_stack(x,filters,kernel_size,strides,conv_shortcut,name,blocks,
	trainable=True):
	""" Build a stack of residual blocks for the xresnet model family

	Args:
		x (KerasTensor): The input Keras tensorflow tensor that will be
			passed through the stack.
		filters (int): The number of output filters
		kernel_size (int): The kernel size of each filter
		strides (int): The strides to use for each convolutional filter
		conv_shortcut (bool): When true, will use a convolutional shortcut and
			when false will use an identity shortcut.
		name (str): The name for this stack
		blocks (int): The number of blocks in this stack
		trainable (bool): If false, weights in this stack will not be trainable.

	Returns:
		(KerasTensor): A Keras tensorflow tensor representing the input after
		the stack has been applied
	"""
	# If the input dimension is not divisible by the stride then we must add
	# padding.
	divide_pad = strides - x.shape[1]%strides
	x = tf.cond(x.shape[1] % strides > 0,lambda:layers.ZeroPadding2D(
		padding=((divide_pad,0),(0,0)),name=name+'_stride_pad_r')(x),
		lambda:x)
	divide_pad = strides - x.shape[2]%strides
	x = tf.cond(x.shape[2] % strides > 0,lambda:layers.ZeroPadding2D(
		padding=((0,0),(divide_pad,0)),name=name+'_stride_pad_c')(x),
		lambda:x)
	# Apply each residual block
	x = _xresnet_block(x,filters,kernel_size,strides,
		conv_shortcut=conv_shortcut,name=name+'_block1',trainable=trainable)
	for i in range(2,blocks+1):
		x = _xresnet_block(x,filters,kernel_size,1,conv_shortcut=False,
			name=name+'_block%d'%(i),trainable=trainable)

	return x


def build_xresnet34(img_size,num_outputs,custom_head=False,
	train_only_head=False):
	""" Build the xresnet34 model described in
	https://arxiv.org/pdf/1812.01187.pdf

	Args:
		img_size ((int,int,int)): A tuple with shape (pix,pix,freq) that
			describes the size of the input images.
		num_outputs (int): The number of outputs to predict
		custom_head (bool): If true, then add a custom head at the end of
			xresnet34 in line with what' used in the fastai code.
		train_only_head (bool): If true, only train the head of the network.

	Returns:
		(keras.Model): An instance of the xresnet34 model implemented in
		Keras.
	"""

	# If we train only the head, then none of the previous weights should be
	# trainable
	trainable = not train_only_head

	# Assume the first dimension is the batch size
	bn_axis = -1

	# Initialize the inputs
	inputs = layers.Input(shape=img_size)

	# Build the stem of the resnet
	# Conv 1 of stem
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name='stem_pad1',
		trainable=trainable)(inputs)
	x = layers.Conv2D(32,3,strides=2,use_bias=False,name='stem_conv1',
		trainable=trainable)(x)
	x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
		name='stem_bn1',trainable=trainable)(x)
	x = layers.Activation('relu',name='stem_relu1',trainable=trainable)(x)

	# Conv 2 of stem
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name='stem_pad2',
		trainable=trainable)(x)
	x = layers.Conv2D(32,3,strides=1,use_bias=False,name='stem_conv2',
		trainable=trainable)(x)
	x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
		name='stem_bn2',trainable=trainable)(x)
	x = layers.Activation('relu',name='stem_relu2',trainable=trainable)(x)

	# Conv 3 of stem
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name='stem_pad3',
		trainable=trainable)(x)
	x = layers.Conv2D(64,3,strides=1,use_bias=False,name='stem_conv3',
		trainable=trainable)(x)
	x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
		name='stem_bn3',trainable=trainable)(x)
	x = layers.Activation('relu',name='stem_relu3',trainable=trainable)(x)

	# Next step is max pooling of the stem
	x = layers.ZeroPadding2D(padding=((1,1),(1,1)),name='stem_pad4',
		trainable=trainable)(x)
	x = layers.MaxPooling2D(3,strides=2,name='stem_maxpooling',
		trainable=trainable)(x)

	# # Now we apply the residual stacks
	x = _xresnet_stack(x,filters=64,kernel_size=3,strides=1,
		conv_shortcut=False,name='stack1',blocks=3,trainable=trainable)
	x = _xresnet_stack(x,filters=128,kernel_size=3,strides=2,
		conv_shortcut=True,name='stack2',blocks=4,trainable=trainable)
	x = _xresnet_stack(x,filters=256,kernel_size=3,strides=2,
		conv_shortcut=True,name='stack3',blocks=6,trainable=trainable)
	x = _xresnet_stack(x,filters=512,kernel_size=3,strides=2,
		conv_shortcut=True,name='stack4',blocks=3,trainable=trainable)

	# Conduct the pooling and a dense transform to the final prediction
	x = layers.GlobalAveragePooling2D(name='avg_pool')(x)
	if custom_head:
		x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
			name='head_bn1')(x)
		x = layers.Dense(512,use_bias=False,name='head_dense1')(x)
		x = layers.Activation('relu',name='head_relu1')(x)
		x = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,momentum=0.1,
			name='head_bn2')(x)
		x = layers.Dense(num_outputs,use_bias=False,name='head_dense2')(x)
		outputs = layers.BatchNormalization(axis=bn_axis,epsilon=1e-5,
			momentum=0.1,name='head_bn3')(x)
	else:
		outputs = layers.Dense(num_outputs,name='output_dense')(x)

	model = Model(inputs=inputs,outputs=outputs)

	return model



# -*- coding: utf-8 -*-
"""
Provides class for specifying a single point source

This module contains the class required to provide a single point source
as the source for paltas.
"""
from .point_source_base import PointSourceBase
from lenstronomy.Util.data_util import magnitude2cps


class SinglePointSource(PointSourceBase):
	"""Class to generate single point source model

	Args:
		point_source_parameters (dict): dictionary with source-specific 
			parameters.

	Notes:

	Required Parameters

	- 	magnitude - AB absolute magnitude of the point source
	-	x_point_source - x-coordinate lens center in units of arcseconds
	-	y_point_source - y-coordinate lens center in units of arcseconds
	- 	z_lens - main deflector redshift
	-	output_ab_zeropoint - AB magnitude zeropoint of the detector
	- 	compute_time_delay - bool determining if time delays will be computed
		and added to the returned kwargs

	Optional Parameters

	- 	mag_pert - list of 4 fractional magnification perturbations that will be
		applied to each point source image
	-	kappa_ext - external convergence used to calculate time delays
	-	time_delay_error - list of errors on the time delay measurements in units
		of days.
	"""

	required_parameters = ('x_point_source', 'y_point_source', 'magnitude',
		'output_ab_zeropoint', 'compute_time_delays')

	def draw_point_source(self):
		"""Return lenstronomy PointSource kwargs

		Returns:
			(list,list) A list containing the model names(s), and
				a list containing the model kwargs dictionaries.
		"""

		point_source_kwargs = {}
		point_source_kwargs['ra_source'] = self.point_source_parameters[
			'x_point_source']
		point_source_kwargs['dec_source'] = self.point_source_parameters[
			'y_point_source']
		if('mag_pert' in self.point_source_parameters.keys()):
			point_source_kwargs['mag_pert'] = self.point_source_parameters[
				'mag_pert']

		# mag to amp conversion
		# note: flux = amplitude for point source
		point_source_kwargs['point_amp'] = magnitude2cps(
			self.point_source_parameters['magnitude'],
			self.point_source_parameters['output_ab_zeropoint'])

		return ['SOURCE_POSITION'], [point_source_kwargs]



# -*- coding: utf-8 -*-
"""
Define the base class to draw line of sight substructure for a lens

This module contains the base class that all the los classes will build
from. Because the steps for rendering halos can vary between different
models, the required functions are very sparse.
"""
from ..Utils.cosmology_utils import get_cosmology
import copy


class LOSBase():
	"""Base class for rendering the los of a main halo.

	Args:
		los_parameters (dict): A dictionary containing the type of
			los distribution and the value for each of its parameters.
		main_deflector_parameters (dict): A dictionary containing the type of
			main deflector and the value for each of its parameters.
		source_parameters (dict): A dictionary containing the type of the
			source and the value for each of its parameters.
		cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
			Either a name of colossus cosmology, a dict with 'cosmology name':
			name of colossus cosmology, an instance of colussus cosmology, or a
			dict with H0 and Om0 ( other parameters will be set to defaults).
	"""

	required_parameters = tuple()

	def __init__(self,los_parameters,main_deflector_parameters,
		source_parameters,cosmology_parameters):

		# Save the parameters as a copy to avoid any misunderstanding on the
		# user end.
		self.los_parameters = copy.deepcopy(los_parameters)
		self.main_deflector_parameters = copy.deepcopy(
			main_deflector_parameters)
		self.source_parameters = copy.deepcopy(source_parameters)

		# Turn our cosmology parameters into a colossus cosmology instance
		self.cosmo = get_cosmology(cosmology_parameters)

		# Check that all the needed parameters are present
		self.check_parameterization(self.__class__.required_parameters)

	def check_parameterization(self,required_params):
		""" Check that all the required parameters are present in the
		los_parameters.

		Args:
			required_params ([str,...]): A list of strings containing the
				required parameters.
		"""
		if not all(elem in self.los_parameters.keys() for
			elem in required_params):
			raise ValueError('Not all of the required parameters for the ' +
				'parameterization are present.')

	def update_parameters(self,los_parameters=None,
		main_deflector_parameters=None,source_parameters=None,
		cosmology_parameters=None):
		"""Updated the class parameters

		Args:
			los_parameters (dict): A dictionary containing the type of
				los distribution and the value for each of its parameters.
			main_deflector_parameters (dict): A dictionary containing the type
				of main deflector and the value for each of its parameters.
			source_parameters (dict): A dictionary containing the type of the
				source and the value for each of its parameters.
			cosmology_parameters (str,dict, or colossus.cosmology.Cosmology):
				Either a name of colossus cosmology, a dict with 'cosmology name':
				name of colossus cosmology, an instance of colussus cosmology, or
				a dict with H0 and Om0 ( other parameters will be set to
				defaults).
		"""
		if los_parameters is not None:
			self.los_parameters = copy.copy(los_parameters)
		if main_deflector_parameters is not None:
			self.main_deflector_parameters = copy.copy(
				main_deflector_parameters)
		if source_parameters is not None:
			self.source_parameters = copy.copy(source_parameters)
		if cosmology_parameters is not None:
			self.cosmo = get_cosmology(cosmology_parameters)

	def draw_los(self):
		"""Draws masses, concentrations,and positions for the los substructure
		of a main lens halo.

		Returns:
			(tuple): A tuple of three lists: the first is the profile type for
			each los halo returned, the second is the lenstronomy kwargs
			for that halo, and the third is a list of redshift values for
			each profile.
		"""
		raise NotImplementedError

	def calculate_average_alpha(self,n_draws=100):
		""" Calculates the average convergence from the los at each redshift
		specified by the los parameters and returns corresponding lenstronomy
		objects.

		Returns:
			(tuple): A tuple of two lists: the first is the interpolation
			profile type for each redshift slice and the second is the
			lenstronomy kwargs for that profile.
		"""
		raise NotImplementedError
